#!/usr/bin/env python

import getopt,sys,subprocess,time,os,re,threading,thread,random,traceback,socket
import pexpect
from xml.dom.minidom import Document
import shlex


# VERSION NUMBER:
MAJOR_VERSION = 0
MINOR_VERSION = 71

# Cluster.conf file location:
CLUSTER_CONF = "/etc/cluster/cluster.conf"

# node list structure:
#  id: #
#  name: str
#  private_ipaddr: str
#  storage_ipaddr: str
#  public_ipaddr: str
#  problems:  [list of str]
#  password: str
#  state: ?

# state values for the node list structure above:
NODE_STATE_INITIAL = 0
NODE_STATE_IDLE = 1
NODE_STATE_RUNNING = 2
NODE_STATE_FAIL = -1

# Some global constants needed for pexepect:
#
PEXPECT_PROMPT = '[PEXPECT]: '
PEXPECT_PROMPT_EXP = re.escape(PEXPECT_PROMPT)
PEXPECT_DRAC_PROMPT_EXP = re.escape('$ ')

ftp_readme_template = """
Temporary file to test FTP server connectivity

To change the FTP configuration, edit the file /etc/vsftpd/vsftpd.conf
on each cluster node with the changes.

The public IP for the FTP server is: %(ipaddr)s
"""

ftp_conf_template = """
# Example config file /etc/vsftpd/vsftpd.conf
#
# The default compiled in settings are fairly paranoid. This sample file
# loosens things up a bit, to make the ftp daemon more usable.
# Please see vsftpd.conf.5 for all compiled in defaults.
#
# READ THIS: This example file is NOT an exhaustive list of vsftpd options.
# Please read the vsftpd.conf.5 manual page to get a full idea of vsftpd's
# capabilities.
#
# Allow anonymous FTP? (Beware - allowed by default if you comment this out).
anonymous_enable=YES
#
# Uncomment this to allow local users to log in.
local_enable=YES
#
# Uncomment this to enable any form of FTP write command.
write_enable=YES
#
# Default umask for local users is 077. You may wish to change this to 022,
# if your users expect that (022 is used by most other ftpd's)
local_umask=022
#
# Uncomment this to allow the anonymous FTP user to upload files. This only
# has an effect if the above global write enable is activated. Also, you will
# obviously need to create a directory writable by the FTP user.
#anon_upload_enable=YES
#
# Uncomment this if you want the anonymous FTP user to be able to create
# new directories.
#anon_mkdir_write_enable=YES
#
# Activate directory messages - messages given to remote users when they
# go into a certain directory.
dirmessage_enable=YES
#
# Activate logging of uploads/downloads.
xferlog_enable=YES
#
# Make sure PORT transfer connections originate from port 20 (ftp-data).
connect_from_port_20=YES
#
# If you want, you can arrange for uploaded anonymous files to be owned by
# a different user. Note! Using "root" for uploaded files is not
# recommended!
#chown_uploads=YES
#chown_username=whoever
#
# You may override where the log file goes if you like. The default is shown
# below.
xferlog_file=/var/log/vsftpd.log
#
# If you want, you can have your log file in standard ftpd xferlog format
xferlog_std_format=YES
#
# You may change the default value for timing out an idle session.
#idle_session_timeout=600
#
# You may change the default value for timing out a data connection.
#data_connection_timeout=120
#
# It is recommended that you define on your system a unique user which the
# ftp server can use as a totally isolated and unprivileged user.
#nopriv_user=ftpsecure
#
# Enable this and the server will recognise asynchronous ABOR requests. Not
# recommended for security (the code is non-trivial). Not enabling it,
# however, may confuse older FTP clients.
#async_abor_enable=YES
#
# By default the server will pretend to allow ASCII mode but in fact ignore
# the request. Turn on the below options to have the server actually do ASCII
# mangling on files when in ASCII mode.
# Beware that on some FTP servers, ASCII support allows a denial of service
# attack (DoS) via the command "SIZE /big/file" in ASCII mode. vsftpd
# predicted this attack and has always been safe, reporting the size of the
# raw file.
# ASCII mangling is a horrible feature of the protocol.
#ascii_upload_enable=YES
#ascii_download_enable=YES
#
# You may fully customise the login banner string:
#ftpd_banner=Welcome to blah FTP service.
#
# You may specify a file of disallowed anonymous e-mail addresses. Apparently
# useful for combatting certain DoS attacks.
#deny_email_enable=YES
# (default follows)
#banned_email_file=/etc/vsftpd/banned_emails
#
# You may specify an explicit list of local users to chroot() to their home
# directory. If chroot_local_user is YES, then this list becomes a list of
# users to NOT chroot().
chroot_list_enable=NO
# (default follows)
#chroot_list_file=/etc/vsftpd/chroot_list
#
# You may activate the "-R" option to the builtin ls. This is disabled by
# default to avoid remote users being able to cause excessive I/O on large
# sites. However, some broken FTP clients such as "ncftp" and "mirror" assume
# the presence of the "-R" option, so there is a strong case for enabling it.
#ls_recurse_enable=YES
#
# When "listen" directive is enabled, vsftpd runs in standalone mode and 
# listens on IPv4 sockets. This directive cannot be used in conjunction 
# with the listen_ipv6 directive.
listen=YES
# Limiting listen to only the cluster IP address for FTP if you want to listen
# to all IP adresses then you will to remove or comment out this line
listen_address=%(ipaddr)s
#
# This directive enables listening on IPv6 sockets. To listen on IPv4 and IPv6
# sockets, you must run two copies of vsftpd whith two configuration files.
# Make sure, that one of the listen options is commented !!
#listen_ipv6=YES

pam_service_name=vsftpd
userlist_enable=YES
tcp_wrappers=YES

anon_root=%(srv_root)s
"""

samba_readme_template = """
Temporary file to test SAMBA server connectivity



To change samba configuration, edit the file /etc/samba/smb.conf.samba-resource
on each cluster node with the changes.

You can use the commands 'findsmb' or 

  smbclient //%(ipaddr)s/pub -N -c "dir"

to see the samba files served out of %(srv_root)s directory


For more information use the Red Hat Administration Guide at:

http://www.redhat.com/docs/manuals/enterprise/RHEL-5-manual/Deployment_Guide-en-US/ch-samba.html

The current configuration is setup for Share-Level Security. Please see 
section 19.7 to change security level if user/password access is desired.

"""

samba_conf_template = """
#
# "/etc/samba/smb.conf.samba-resource"
#
# This template configuration wass automatically generated, and will
# be automatically regenerated if removed.  Please modify this file to
# speficy subdirectories and/or client access permissions.
#
# Once this file has been altered, automatic re-generation will stop.
# Remember to copy this file to all other cluster members after making
# changes, or your SMB service will not operate correctly.
#
# From a cluster perspective, the key fields are:
#     lock directory - must be unique per samba service.
#     bind interfaces only - must be present set to yes.
#     interfaces - must be set to service floating IP address.
#     path - must be the service mountpoint or subdirectory thereof.
#


[global]
        workgroup = %(workgroup)s
        pid directory = /var/run/samba/samba-resource
        lock directory = /var/cache/samba/samba-resource
	log file = /var/log/samba/%%m.log
	#private dir = /var/
	encrypt passwords = yes
	bind interfaces only = yes
	netbios name = samba-resource
        security = share

[pub]
        comment = The public stuff
        path = %(srv_root)s
        public = yes
        writable = yes
        browseable = yes
        read only = yes
        guest ok = yes
        force user = nobody
        force group = nobody
       

	#
	# Interfaces are based on ip resources at the top level of
	# "samba-srv"; IPv6 addresses may or may not
	# work correctly.
	#
	interfaces = %(ipaddr)s

"""

http_readme_template = """
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">

<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">
	<head>
		<title>Test Page for the Apache HTTP Server on Red Hat Enterprise Linux</title>
		<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
		<style type="text/css">
			/*<![CDATA[*/
			body {
				background-color: #fff;
				color: #000;
				font-size: 0.9em;
				font-family: sans-serif,helvetica;
				margin: 0;
				padding: 0;
			}
			:link {
				color: #c00;
			}
			:visited {
				color: #c00;
			}
			a:hover {
				color: #f50;
			}
			h1 {
				text-align: center;
				margin: 0;
				padding: 0.6em 2em 0.4em;
				background-color: #900;
				color: #fff;
				font-weight: normal;
				font-size: 1.75em;
				border-bottom: 2px solid #000;
			}
			h1 strong {
				font-weight: bold;
			}
			h2 {
				font-size: 1.1em;
				font-weight: bold;
			}
			hr {
				display: none;
			}
			.content {
				padding: 1em 5em;
			}
			.content-columns {
				/* Setting relative positioning allows for 
				absolute positioning for sub-classes */
				position: relative;
				padding-top: 1em;
			}
			.content-column-left {
				/* Value for IE/Win; will be overwritten for other browsers */
				width: 47%%;
				padding-right: 3%%;
				float: left;
				padding-bottom: 2em;
			}
			.content-column-left hr {
				display: none;
			}
			.content-column-right {
				/* Values for IE/Win; will be overwritten for other browsers */
				width: 47%%;
				padding-left: 3%%;
				float: left;
				padding-bottom: 2em;
			}
			.content-columns>.content-column-left, .content-columns>.content-column-right {
				/* Non-IE/Win */
			}
			img {
				border: 2px solid #fff;
				padding: 2px;
				margin: 2px;
			}
			a:hover img {
				border: 2px solid #f50;
			}
			/*]]>*/
		</style>
	</head>

	<body>
		<h1>Red Hat Enterprise Linux <strong>Test Page</strong></h1>

		<div class="content">
			<div class="content-middle">
				<p>This page is used to test the proper operation of the Apache HTTP server after it has been installed. If you can read this page, it means that the Apache HTTP server installed at this site is working properly.</p>
			</div>
			<hr />

			<div class="content-columns">
				<div class="content-column-left">
					<h2>If you are a member of the general public:</h2>

					<p>The fact that you are seeing this page indicates that the website you just visited is either experiencing problems, or is undergoing routine maintenance.</p>

					<p>If you would like to let the administrators of this website know that you've seen this page instead of the page you expected, you should send them e-mail. In general, mail sent to the name "webmaster" and directed to the website's domain should reach the appropriate person.</p>

					<p>For example, if you experienced problems while visiting www.example.com, you should send e-mail to "webmaster@example.com".</p>

					<p>For information on Red Hat Enterprise Linux, please visit the <a href="http://www.redhat.com/">Red Hat, Inc. website</a>. The documentation for Red Hat Enterprise Linux is <a href="http://www.redhat.com/docs/manuals/enterprise/">available on the Red Hat, Inc. website</a>.</p>
					<hr />
				</div>

				<div class="content-column-right">
					<h2>If you are the website administrator:</h2>

					<p>You may now add content to the directory <tt>%(srv_root)s/html/</tt>. Note that until you do so, people visiting your website will see this page, and not your content. To prevent this page from ever being used, follow the instructions in the file <tt>%(srv_root)s/conf.d/welcome.conf</tt>.</p>

                                        <p>You are free to use the image below on web sites powered by the Apache HTTP Server:</p>
					
                                        <p align="center"><a href="http://httpd.apache.org/"><img src="/icons/apache_pb2.gif" alt="[ Powered by Apache ]"/></a></p>

				</div>
			</div>
		</div>
	</body>
</html>
"""

http_conf_template = """
#
# This is the main Apache server configuration file.  It contains the
# configuration directives that give the server its instructions.
# See <URL:http://httpd.apache.org/docs/2.2/> for detailed information.
# In particular, see
# <URL:http://httpd.apache.org/docs/2.2/mod/directives.html>
# for a discussion of each configuration directive.
#
#
# Do NOT simply read the instructions in here without understanding
# what they do.  They're here only as hints or reminders.  If you are unsure
# consult the online docs. You have been warned.  
#
# The configuration directives are grouped into three basic sections:
#  1. Directives that control the operation of the Apache server process as a
#     whole (the 'global environment').
#  2. Directives that define the parameters of the 'main' or 'default' server,
#     which responds to requests that aren't handled by a virtual host.
#     These directives also provide default values for the settings
#     of all virtual hosts.
#  3. Settings for virtual hosts, which allow Web requests to be sent to
#     different IP addresses or hostnames and have them handled by the
#     same Apache server process.
#
# Configuration and logfile names: If the filenames you specify for many
# of the server's control files begin with "/" (or "drive:/" for Win32), the
# server will use that explicit path.  If the filenames do *not* begin
# with "/", the value of ServerRoot is prepended -- so "logs/foo.log"
# with ServerRoot set to "/etc/httpd" will be interpreted by the
# server as "/etc/httpd/logs/foo.log".
#

### Section 1: Global Environment
#
# The directives in this section affect the overall operation of Apache,
# such as the number of concurrent requests it can handle or where it
# can find its configuration files.
#

#
# Don't give away too much information about all the subcomponents
# we are running.  Comment out this line if you don't mind remote sites
# finding out what major optional modules you are running
ServerTokens OS

#
# ServerRoot: The top of the directory tree under which the server's
# configuration, error, and log files are kept.
#
# NOTE!  If you intend to place this on an NFS (or otherwise network)
# mounted filesystem then please read the LockFile documentation
# (available at <URL:http://httpd.apache.org/docs/2.2/mod/mpm_common.html#lockfile>);
# you will save yourself a lot of trouble.
#
# Do NOT add a slash at the end of the directory path.
#
ServerRoot "%(srv_root)s"

#
# PidFile: The file in which the server should record its process
# identification number when it starts.
#
PidFile run/httpd.pid

#
# Timeout: The number of seconds before receives and sends time out.
#
Timeout 120

#
# KeepAlive: Whether or not to allow persistent connections (more than
# one request per connection). Set to "Off" to deactivate.
#
KeepAlive Off

#
# MaxKeepAliveRequests: The maximum number of requests to allow
# during a persistent connection. Set to 0 to allow an unlimited amount.
# We recommend you leave this number high, for maximum performance.
#
MaxKeepAliveRequests 100

#
# KeepAliveTimeout: Number of seconds to wait for the next request from the
# same client on the same connection.
#
KeepAliveTimeout 15

##
## Server-Pool Size Regulation (MPM specific)
## 

# prefork MPM
# StartServers: number of server processes to start
# MinSpareServers: minimum number of server processes which are kept spare
# MaxSpareServers: maximum number of server processes which are kept spare
# ServerLimit: maximum value for MaxClients for the lifetime of the server
# MaxClients: maximum number of server processes allowed to start
# MaxRequestsPerChild: maximum number of requests a server process serves
<IfModule prefork.c>
StartServers       8
MinSpareServers    5
MaxSpareServers   20
ServerLimit      256
MaxClients       256
MaxRequestsPerChild  4000
</IfModule>

# worker MPM
# StartServers: initial number of server processes to start
# MaxClients: maximum number of simultaneous client connections
# MinSpareThreads: minimum number of worker threads which are kept spare
# MaxSpareThreads: maximum number of worker threads which are kept spare
# ThreadsPerChild: constant number of worker threads in each server process
# MaxRequestsPerChild: maximum number of requests a server process serves
<IfModule worker.c>
StartServers         2
MaxClients         150
MinSpareThreads     25
MaxSpareThreads     75 
ThreadsPerChild     25
MaxRequestsPerChild  0
</IfModule>

#
# Listen: Allows you to bind Apache to specific IP addresses and/or
# ports, in addition to the default. See also the <VirtualHost>
# directive.
#
# Change this to Listen on specific IP addresses as shown below to 
# prevent Apache from glomming onto all bound IP addresses (0.0.0.0)
#
#Listen 12.34.56.78:80
Listen %(ipaddr)s:80

#
# Dynamic Shared Object (DSO) Support
#
# To be able to use the functionality of a module which was built as a DSO you
# have to place corresponding `LoadModule' lines at this location so the
# directives contained in it are actually available _before_ they are used.
# Statically compiled modules (those listed by `httpd -l') do not need
# to be loaded here.
#
# Example:
# LoadModule foo_module modules/mod_foo.so
#
LoadModule auth_basic_module modules/mod_auth_basic.so
LoadModule auth_digest_module modules/mod_auth_digest.so
LoadModule authn_file_module modules/mod_authn_file.so
LoadModule authn_alias_module modules/mod_authn_alias.so
LoadModule authn_anon_module modules/mod_authn_anon.so
LoadModule authn_dbm_module modules/mod_authn_dbm.so
LoadModule authn_default_module modules/mod_authn_default.so
LoadModule authz_host_module modules/mod_authz_host.so
LoadModule authz_user_module modules/mod_authz_user.so
LoadModule authz_owner_module modules/mod_authz_owner.so
LoadModule authz_groupfile_module modules/mod_authz_groupfile.so
LoadModule authz_dbm_module modules/mod_authz_dbm.so
LoadModule authz_default_module modules/mod_authz_default.so
LoadModule ldap_module modules/mod_ldap.so
LoadModule authnz_ldap_module modules/mod_authnz_ldap.so
LoadModule include_module modules/mod_include.so
LoadModule log_config_module modules/mod_log_config.so
LoadModule logio_module modules/mod_logio.so
LoadModule env_module modules/mod_env.so
LoadModule ext_filter_module modules/mod_ext_filter.so
LoadModule mime_magic_module modules/mod_mime_magic.so
LoadModule expires_module modules/mod_expires.so
LoadModule deflate_module modules/mod_deflate.so
LoadModule headers_module modules/mod_headers.so
LoadModule usertrack_module modules/mod_usertrack.so
LoadModule setenvif_module modules/mod_setenvif.so
LoadModule mime_module modules/mod_mime.so
LoadModule dav_module modules/mod_dav.so
LoadModule status_module modules/mod_status.so
LoadModule autoindex_module modules/mod_autoindex.so
LoadModule info_module modules/mod_info.so
LoadModule dav_fs_module modules/mod_dav_fs.so
LoadModule vhost_alias_module modules/mod_vhost_alias.so
LoadModule negotiation_module modules/mod_negotiation.so
LoadModule dir_module modules/mod_dir.so
LoadModule actions_module modules/mod_actions.so
LoadModule speling_module modules/mod_speling.so
LoadModule userdir_module modules/mod_userdir.so
LoadModule alias_module modules/mod_alias.so
LoadModule rewrite_module modules/mod_rewrite.so
LoadModule proxy_module modules/mod_proxy.so
LoadModule proxy_balancer_module modules/mod_proxy_balancer.so
LoadModule proxy_ftp_module modules/mod_proxy_ftp.so
LoadModule proxy_http_module modules/mod_proxy_http.so
LoadModule proxy_connect_module modules/mod_proxy_connect.so
LoadModule cache_module modules/mod_cache.so
LoadModule suexec_module modules/mod_suexec.so
LoadModule disk_cache_module modules/mod_disk_cache.so
LoadModule file_cache_module modules/mod_file_cache.so
LoadModule mem_cache_module modules/mod_mem_cache.so
LoadModule cgi_module modules/mod_cgi.so
LoadModule version_module modules/mod_version.so

#
# The following modules are not loaded by default:
#
#LoadModule cern_meta_module modules/mod_cern_meta.so
#LoadModule asis_module modules/mod_asis.so

#
# Load config files from the config directory "/etc/httpd/conf.d".
#
Include conf.d/*.conf

#
# ExtendedStatus controls whether Apache will generate "full" status
# information (ExtendedStatus On) or just basic information (ExtendedStatus
# Off) when the "server-status" handler is called. The default is Off.
#
#ExtendedStatus On

#
# If you wish httpd to run as a different user or group, you must run
# httpd as root initially and it will switch.  
#
# User/Group: The name (or #number) of the user/group to run httpd as.
#  . On SCO (ODT 3) use "User nouser" and "Group nogroup".
#  . On HPUX you may not be able to use shared memory as nobody, and the
#    suggested workaround is to create a user www and use that user.
#  NOTE that some kernels refuse to setgid(Group) or semctl(IPC_SET)
#  when the value of (unsigned)Group is above 60000; 
#  don't use Group #-1 on these systems!
#
User apache
Group apache

### Section 2: 'Main' server configuration
#
# The directives in this section set up the values used by the 'main'
# server, which responds to any requests that aren't handled by a
# <VirtualHost> definition.  These values also provide defaults for
# any <VirtualHost> containers you may define later in the file.
#
# All of these directives may appear inside <VirtualHost> containers,
# in which case these default settings will be overridden for the
# virtual host being defined.
#

#
# ServerAdmin: Your address, where problems with the server should be
# e-mailed.  This address appears on some server-generated pages, such
# as error documents.  e.g. admin@your-domain.com
#
ServerAdmin root@localhost

#
# ServerName gives the name and port that the server uses to identify itself.
# This can often be determined automatically, but we recommend you specify
# it explicitly to prevent problems during startup.
#
# If this is not set to valid DNS name for your host, server-generated
# redirections will not work.  See also the UseCanonicalName directive.
#
# If your host doesn't have a registered DNS name, enter its IP address here.
# You will have to access it by its address anyway, and this will make 
# redirections work in a sensible way.
#
#ServerName www.example.com:80

#
# UseCanonicalName: Determines how Apache constructs self-referencing 
# URLs and the SERVER_NAME and SERVER_PORT variables.
# When set "Off", Apache will use the Hostname and Port supplied
# by the client.  When set "On", Apache will use the value of the
# ServerName directive.
#
UseCanonicalName Off

#
# DocumentRoot: The directory out of which you will serve your
# documents. By default, all requests are taken from this directory, but
# symbolic links and aliases may be used to point to other locations.
#
DocumentRoot "%(doc_root)s"

#
# Each directory to which Apache has access can be configured with respect
# to which services and features are allowed and/or disabled in that
# directory (and its subdirectories). 
#
# First, we configure the "default" to be a very restrictive set of 
# features.  
#
<Directory />
    Options FollowSymLinks
    AllowOverride None
</Directory>

#
# Note that from this point forward you must specifically allow
# particular features to be enabled - so if something's not working as
# you might expect, make sure that you have specifically enabled it
# below.
#

#
# This should be changed to whatever you set DocumentRoot to.
#
<Directory "%(doc_root)s">

#
# Possible values for the Options directive are "None", "All",
# or any combination of:
#   Indexes Includes FollowSymLinks SymLinksifOwnerMatch ExecCGI MultiViews
#
# Note that "MultiViews" must be named *explicitly* --- "Options All"
# doesn't give it to you.
#
# The Options directive is both complicated and important.  Please see
# http://httpd.apache.org/docs/2.2/mod/core.html#options
# for more information.
#
    Options Indexes FollowSymLinks

#
# AllowOverride controls what directives may be placed in .htaccess files.
# It can be "All", "None", or any combination of the keywords:
#   Options FileInfo AuthConfig Limit
#
    AllowOverride None

#
# Controls who can get stuff from this server.
#
    Order allow,deny
    Allow from all

</Directory>

#
# UserDir: The name of the directory that is appended onto a user's home
# directory if a ~user request is received.
#
# The path to the end user account 'public_html' directory must be
# accessible to the webserver userid.  This usually means that ~userid
# must have permissions of 711, ~userid/public_html must have permissions
# of 755, and documents contained therein must be world-readable.
# Otherwise, the client will only receive a "403 Forbidden" message.
#
# See also: http://httpd.apache.org/docs/misc/FAQ.html#forbidden
#
<IfModule mod_userdir.c>
    #
    # UserDir is disabled by default since it can confirm the presence
    # of a username on the system (depending on home directory
    # permissions).
    #
    UserDir disable

    #
    # To enable requests to /~user/ to serve the user's public_html
    # directory, remove the "UserDir disable" line above, and uncomment
    # the following line instead:
    # 
    #UserDir public_html

</IfModule>

#
# Control access to UserDir directories.  The following is an example
# for a site where these directories are restricted to read-only.
#
#<Directory /home/*/public_html>
#    AllowOverride FileInfo AuthConfig Limit
#    Options MultiViews Indexes SymLinksIfOwnerMatch IncludesNoExec
#    <Limit GET POST OPTIONS>
#        Order allow,deny
#        Allow from all
#    </Limit>
#    <LimitExcept GET POST OPTIONS>
#        Order deny,allow
#        Deny from all
#    </LimitExcept>
#</Directory>

#
# DirectoryIndex: sets the file that Apache will serve if a directory
# is requested.
#
# The index.html.var file (a type-map) is used to deliver content-
# negotiated documents.  The MultiViews Option can be used for the 
# same purpose, but it is much slower.
#
DirectoryIndex index.html index.html.var

#
# AccessFileName: The name of the file to look for in each directory
# for additional configuration directives.  See also the AllowOverride
# directive.
#
AccessFileName .htaccess

#
# The following lines prevent .htaccess and .htpasswd files from being 
# viewed by Web clients. 
#
<Files ~ "^\.ht">
    Order allow,deny
    Deny from all
</Files>

#
# TypesConfig describes where the mime.types file (or equivalent) is
# to be found.
#
TypesConfig /etc/mime.types

#
# DefaultType is the default MIME type the server will use for a document
# if it cannot otherwise determine one, such as from filename extensions.
# If your server contains mostly text or HTML documents, "text/plain" is
# a good value.  If most of your content is binary, such as applications
# or images, you may want to use "application/octet-stream" instead to
# keep browsers from trying to display binary files as though they are
# text.
#
DefaultType text/plain

#
# The mod_mime_magic module allows the server to use various hints from the
# contents of the file itself to determine its type.  The MIMEMagicFile
# directive tells the module where the hint definitions are located.
#
<IfModule mod_mime_magic.c>
#   MIMEMagicFile /usr/share/magic.mime
    MIMEMagicFile conf/magic
</IfModule>

#
# HostnameLookups: Log the names of clients or just their IP addresses
# e.g., www.apache.org (on) or 204.62.129.132 (off).
# The default is off because it'd be overall better for the net if people
# had to knowingly turn this feature on, since enabling it means that
# each client request will result in AT LEAST one lookup request to the
# nameserver.
#
HostnameLookups Off

#
# EnableMMAP: Control whether memory-mapping is used to deliver
# files (assuming that the underlying OS supports it).
# The default is on; turn this off if you serve from NFS-mounted 
# filesystems.  On some systems, turning it off (regardless of
# filesystem) can improve performance; for details, please see
# http://httpd.apache.org/docs/2.2/mod/core.html#enablemmap
#
#EnableMMAP off

#
# EnableSendfile: Control whether the sendfile kernel support is 
# used to deliver files (assuming that the OS supports it). 
# The default is on; turn this off if you serve from NFS-mounted 
# filesystems.  Please see
# http://httpd.apache.org/docs/2.2/mod/core.html#enablesendfile
#
#EnableSendfile off

#
# ErrorLog: The location of the error log file.
# If you do not specify an ErrorLog directive within a <VirtualHost>
# container, error messages relating to that virtual host will be
# logged here.  If you *do* define an error logfile for a <VirtualHost>
# container, that host's errors will be logged there and not here.
#
ErrorLog logs/error_log

#
# LogLevel: Control the number of messages logged to the error_log.
# Possible values include: debug, info, notice, warn, error, crit,
# alert, emerg.
#
LogLevel warn

#
# The following directives define some format nicknames for use with
# a CustomLog directive (see below).
#
LogFormat "%%h %%l %%u %%t \\"%%r\\" %%>s %%b \\"%%{Referer}i\\" \\"%%{User-Agent}i\\"" combined
LogFormat "%%h %%l %%u %%t \\"%%r\\" %%>s %%b" common
LogFormat "%%{Referer}i -> %%U" referer
LogFormat "%%{User-agent}i" agent

# "combinedio" includes actual counts of actual bytes received (%%I) and sent (%%O); this
# requires the mod_logio module to be loaded.
#LogFormat "%%h %%l %%u %%t \\"%%r\\" %%>s %%b \\"%%{Referer}i\\" \\"%%{User-Agent}i\\" %%I %%O" combinedio

#
# The location and format of the access logfile (Common Logfile Format).
# If you do not define any access logfiles within a <VirtualHost>
# container, they will be logged here.  Contrariwise, if you *do*
# define per-<VirtualHost> access logfiles, transactions will be
# logged therein and *not* in this file.
#
#CustomLog logs/access_log common

#
# If you would like to have separate agent and referer logfiles, uncomment
# the following directives.
#
#CustomLog logs/referer_log referer
#CustomLog logs/agent_log agent

#
# For a single logfile with access, agent, and referer information
# (Combined Logfile Format), use the following directive:
#
CustomLog logs/access_log combined

#
# Optionally add a line containing the server version and virtual host
# name to server-generated pages (internal error documents, FTP directory
# listings, mod_status and mod_info output etc., but not CGI generated
# documents or custom error documents).
# Set to "EMail" to also include a mailto: link to the ServerAdmin.
# Set to one of:  On | Off | EMail
#
ServerSignature On

#
# Aliases: Add here as many aliases as you need (with no limit). The format is 
# Alias fakename realname
#
# Note that if you include a trailing / on fakename then the server will
# require it to be present in the URL.  So "/icons" isn't aliased in this
# example, only "/icons/".  If the fakename is slash-terminated, then the 
# realname must also be slash terminated, and if the fakename omits the 
# trailing slash, the realname must also omit it.
#
# We include the /icons/ alias for FancyIndexed directory listings.  If you
# do not use FancyIndexing, you may comment this out.
#
Alias /icons/ "%(srv_root)s/icons/"

<Directory "%(srv_root)s/icons">
    Options Indexes MultiViews
    AllowOverride None
    Order allow,deny
    Allow from all
</Directory>

#
# WebDAV module configuration section.
# 
<IfModule mod_dav_fs.c>
    # Location of the WebDAV lock database.
    DAVLockDB /var/lib/dav/lockdb
</IfModule>

#
# ScriptAlias: This controls which directories contain server scripts.
# ScriptAliases are essentially the same as Aliases, except that
# documents in the realname directory are treated as applications and
# run by the server when requested rather than as documents sent to the client.
# The same rules about trailing "/" apply to ScriptAlias directives as to
# Alias.
#
ScriptAlias /cgi-bin/ "%(srv_root)s/cgi-bin/"

#
# "/var/www/cgi-bin" should be changed to whatever your ScriptAliased
# CGI directory exists, if you have that configured.
#
<Directory "%(srv_root)s/cgi-bin">
    AllowOverride None
    Options None
    Order allow,deny
    Allow from all
</Directory>

#
# Redirect allows you to tell clients about documents which used to exist in
# your server's namespace, but do not anymore. This allows you to tell the
# clients where to look for the relocated document.
# Example:
# Redirect permanent /foo http://www.example.com/bar

#
# Directives controlling the display of server-generated directory listings.
#

#
# IndexOptions: Controls the appearance of server-generated directory
# listings.
#
IndexOptions FancyIndexing VersionSort NameWidth=* HTMLTable

#
# AddIcon* directives tell the server which icon to show for different
# files or filename extensions.  These are only displayed for
# FancyIndexed directories.
#
AddIconByEncoding (CMP,/icons/compressed.gif) x-compress x-gzip

AddIconByType (TXT,/icons/text.gif) text/*
AddIconByType (IMG,/icons/image2.gif) image/*
AddIconByType (SND,/icons/sound2.gif) audio/*
AddIconByType (VID,/icons/movie.gif) video/*

AddIcon /icons/binary.gif .bin .exe
AddIcon /icons/binhex.gif .hqx
AddIcon /icons/tar.gif .tar
AddIcon /icons/world2.gif .wrl .wrl.gz .vrml .vrm .iv
AddIcon /icons/compressed.gif .Z .z .tgz .gz .zip
AddIcon /icons/a.gif .ps .ai .eps
AddIcon /icons/layout.gif .html .shtml .htm .pdf
AddIcon /icons/text.gif .txt
AddIcon /icons/c.gif .c
AddIcon /icons/p.gif .pl .py
AddIcon /icons/f.gif .for
AddIcon /icons/dvi.gif .dvi
AddIcon /icons/uuencoded.gif .uu
AddIcon /icons/script.gif .conf .sh .shar .csh .ksh .tcl
AddIcon /icons/tex.gif .tex
AddIcon /icons/bomb.gif core

AddIcon /icons/back.gif ..
AddIcon /icons/hand.right.gif README
AddIcon /icons/folder.gif ^^DIRECTORY^^
AddIcon /icons/blank.gif ^^BLANKICON^^

#
# DefaultIcon is which icon to show for files which do not have an icon
# explicitly set.
#
DefaultIcon /icons/unknown.gif

#
# AddDescription allows you to place a short description after a file in
# server-generated indexes.  These are only displayed for FancyIndexed
# directories.
# Format: AddDescription "description" filename
#
#AddDescription "GZIP compressed document" .gz
#AddDescription "tar archive" .tar
#AddDescription "GZIP compressed tar archive" .tgz

#
# ReadmeName is the name of the README file the server will look for by
# default, and append to directory listings.
#
# HeaderName is the name of a file which should be prepended to
# directory indexes. 
ReadmeName README.html
HeaderName HEADER.html

#
# IndexIgnore is a set of filenames which directory indexing should ignore
# and not include in the listing.  Shell-style wildcarding is permitted.
#
IndexIgnore .??* *~ *# HEADER* README* RCS CVS *,v *,t

#
# DefaultLanguage and AddLanguage allows you to specify the language of 
# a document. You can then use content negotiation to give a browser a 
# file in a language the user can understand.
#
# Specify a default language. This means that all data
# going out without a specific language tag (see below) will 
# be marked with this one. You probably do NOT want to set
# this unless you are sure it is correct for all cases.
#
# * It is generally better to not mark a page as 
# * being a certain language than marking it with the wrong
# * language!
#
# DefaultLanguage nl
#
# Note 1: The suffix does not have to be the same as the language
# keyword --- those with documents in Polish (whose net-standard
# language code is pl) may wish to use "AddLanguage pl .po" to
# avoid the ambiguity with the common suffix for perl scripts.
#
# Note 2: The example entries below illustrate that in some cases 
# the two character 'Language' abbreviation is not identical to 
# the two character 'Country' code for its country,
# E.g. 'Danmark/dk' versus 'Danish/da'.
#
# Note 3: In the case of 'ltz' we violate the RFC by using a three char
# specifier. There is 'work in progress' to fix this and get
# the reference data for rfc1766 cleaned up.
#
# Catalan (ca) - Croatian (hr) - Czech (cs) - Danish (da) - Dutch (nl)
# English (en) - Esperanto (eo) - Estonian (et) - French (fr) - German (de)
# Greek-Modern (el) - Hebrew (he) - Italian (it) - Japanese (ja)
# Korean (ko) - Luxembourgeois* (ltz) - Norwegian Nynorsk (nn)
# Norwegian (no) - Polish (pl) - Portugese (pt)
# Brazilian Portuguese (pt-BR) - Russian (ru) - Swedish (sv)
# Simplified Chinese (zh-CN) - Spanish (es) - Traditional Chinese (zh-TW)
#
AddLanguage ca .ca
AddLanguage cs .cz .cs
AddLanguage da .dk
AddLanguage de .de
AddLanguage el .el
AddLanguage en .en
AddLanguage eo .eo
AddLanguage es .es
AddLanguage et .et
AddLanguage fr .fr
AddLanguage he .he
AddLanguage hr .hr
AddLanguage it .it
AddLanguage ja .ja
AddLanguage ko .ko
AddLanguage ltz .ltz
AddLanguage nl .nl
AddLanguage nn .nn
AddLanguage no .no
AddLanguage pl .po
AddLanguage pt .pt
AddLanguage pt-BR .pt-br
AddLanguage ru .ru
AddLanguage sv .sv
AddLanguage zh-CN .zh-cn
AddLanguage zh-TW .zh-tw

#
# LanguagePriority allows you to give precedence to some languages
# in case of a tie during content negotiation.
#
# Just list the languages in decreasing order of preference. We have
# more or less alphabetized them here. You probably want to change this.
#
LanguagePriority en ca cs da de el eo es et fr he hr it ja ko ltz nl nn no pl pt pt-BR ru sv zh-CN zh-TW

#
# ForceLanguagePriority allows you to serve a result page rather than
# MULTIPLE CHOICES (Prefer) [in case of a tie] or NOT ACCEPTABLE (Fallback)
# [in case no accepted languages matched the available variants]
#
ForceLanguagePriority Prefer Fallback

#
# Specify a default charset for all content served; this enables
# interpretation of all content as UTF-8 by default.  To use the 
# default browser choice (ISO-8859-1), or to allow the META tags
# in HTML content to override this choice, comment out this
# directive:
#
AddDefaultCharset UTF-8

#
# AddType allows you to add to or override the MIME configuration
# file mime.types for specific file types.
#
#AddType application/x-tar .tgz

#
# AddEncoding allows you to have certain browsers uncompress
# information on the fly. Note: Not all browsers support this.
# Despite the name similarity, the following Add* directives have nothing
# to do with the FancyIndexing customization directives above.
#
#AddEncoding x-compress .Z
#AddEncoding x-gzip .gz .tgz

# If the AddEncoding directives above are commented-out, then you
# probably should define those extensions to indicate media types:
#
AddType application/x-compress .Z
AddType application/x-gzip .gz .tgz

#
# AddHandler allows you to map certain file extensions to "handlers":
# actions unrelated to filetype. These can be either built into the server
# or added with the Action directive (see below)
#
# To use CGI scripts outside of ScriptAliased directories:
# (You will also need to add "ExecCGI" to the "Options" directive.)
#
#AddHandler cgi-script .cgi

#
# For files that include their own HTTP headers:
#
#AddHandler send-as-is asis

#
# For type maps (negotiated resources):
# (This is enabled by default to allow the Apache "It Worked" page
#  to be distributed in multiple languages.)
#
AddHandler type-map var

#
# Filters allow you to process content before it is sent to the client.
#
# To parse .shtml files for server-side includes (SSI):
# (You will also need to add "Includes" to the "Options" directive.)
#
AddType text/html .shtml
AddOutputFilter INCLUDES .shtml

#
# Action lets you define media types that will execute a script whenever
# a matching file is called. This eliminates the need for repeated URL
# pathnames for oft-used CGI file processors.
# Format: Action media/type /cgi-script/location
# Format: Action handler-name /cgi-script/location
#

#
# Customizable error responses come in three flavors:
# 1) plain text 2) local redirects 3) external redirects
#
# Some examples:
#ErrorDocument 500 "The server made a boo boo."
#ErrorDocument 404 /missing.html
#ErrorDocument 404 "/cgi-bin/missing_handler.pl"
#ErrorDocument 402 http://www.example.com/subscription_info.html
#

#
# Putting this all together, we can internationalize error responses.
#
# We use Alias to redirect any /error/HTTP_<error>.html.var response to
# our collection of by-error message multi-language collections.  We use 
# includes to substitute the appropriate text.
#
# You can modify the messages' appearance without changing any of the
# default HTTP_<error>.html.var files by adding the line:
#
#   Alias /error/include/ "/your/include/path/"
#
# which allows you to create your own set of files by starting with the
# /var/www/error/include/ files and
# copying them to /your/include/path/, even on a per-VirtualHost basis.
#

Alias /error/ "%(srv_root)s/error/"

<IfModule mod_negotiation.c>
<IfModule mod_include.c>
    <Directory "%(srv_root)s/error">
        AllowOverride None
        Options IncludesNoExec
        AddOutputFilter Includes html
        AddHandler type-map var
        Order allow,deny
        Allow from all
        LanguagePriority en es de fr
        ForceLanguagePriority Prefer Fallback
    </Directory>

#    ErrorDocument 400 /error/HTTP_BAD_REQUEST.html.var
#    ErrorDocument 401 /error/HTTP_UNAUTHORIZED.html.var
#    ErrorDocument 403 /error/HTTP_FORBIDDEN.html.var
#    ErrorDocument 404 /error/HTTP_NOT_FOUND.html.var
#    ErrorDocument 405 /error/HTTP_METHOD_NOT_ALLOWED.html.var
#    ErrorDocument 408 /error/HTTP_REQUEST_TIME_OUT.html.var
#    ErrorDocument 410 /error/HTTP_GONE.html.var
#    ErrorDocument 411 /error/HTTP_LENGTH_REQUIRED.html.var
#    ErrorDocument 412 /error/HTTP_PRECONDITION_FAILED.html.var
#    ErrorDocument 413 /error/HTTP_REQUEST_ENTITY_TOO_LARGE.html.var
#    ErrorDocument 414 /error/HTTP_REQUEST_URI_TOO_LARGE.html.var
#    ErrorDocument 415 /error/HTTP_UNSUPPORTED_MEDIA_TYPE.html.var
#    ErrorDocument 500 /error/HTTP_INTERNAL_SERVER_ERROR.html.var
#    ErrorDocument 501 /error/HTTP_NOT_IMPLEMENTED.html.var
#    ErrorDocument 502 /error/HTTP_BAD_GATEWAY.html.var
#    ErrorDocument 503 /error/HTTP_SERVICE_UNAVAILABLE.html.var
#    ErrorDocument 506 /error/HTTP_VARIANT_ALSO_VARIES.html.var

</IfModule>
</IfModule>

#
# The following directives modify normal HTTP response behavior to
# handle known problems with browser implementations.
#
BrowserMatch "Mozilla/2" nokeepalive
BrowserMatch "MSIE 4\.0b2;" nokeepalive downgrade-1.0 force-response-1.0
BrowserMatch "RealPlayer 4\.0" force-response-1.0
BrowserMatch "Java/1\.0" force-response-1.0
BrowserMatch "JDK/1\.0" force-response-1.0

#
# The following directive disables redirects on non-GET requests for
# a directory that does not include the trailing slash.  This fixes a 
# problem with Microsoft WebFolders which does not appropriately handle 
# redirects for folders with DAV methods.
# Same deal with Apple's DAV filesystem and Gnome VFS support for DAV.
#
BrowserMatch "Microsoft Data Access Internet Publishing Provider" redirect-carefully
BrowserMatch "MS FrontPage" redirect-carefully
BrowserMatch "^WebDrive" redirect-carefully
BrowserMatch "^WebDAVFS/1.[0123]" redirect-carefully
BrowserMatch "^gnome-vfs/1.0" redirect-carefully
BrowserMatch "^XML Spy" redirect-carefully
BrowserMatch "^Dreamweaver-WebDAV-SCM1" redirect-carefully

#
# Allow server status reports generated by mod_status,
# with the URL of http://servername/server-status
# Change the ".example.com" to match your domain to enable.
#
#<Location /server-status>
#    SetHandler server-status
#    Order deny,allow
#    Deny from all
#    Allow from .example.com
#</Location>

#
# Allow remote server configuration reports, with the URL of
#  http://servername/server-info (requires that mod_info.c be loaded).
# Change the ".example.com" to match your domain to enable.
#
#<Location /server-info>
#    SetHandler server-info
#    Order deny,allow
#    Deny from all
#    Allow from .example.com
#</Location>

#
# Proxy Server directives. Uncomment the following lines to
# enable the proxy server:
#
#<IfModule mod_proxy.c>
#ProxyRequests On
#
#<Proxy *>
#    Order deny,allow
#    Deny from all
#    Allow from .example.com
#</Proxy>

#
# Enable/disable the handling of HTTP/1.1 "Via:" headers.
# ("Full" adds the server version; "Block" removes all outgoing Via: headers)
# Set to one of: Off | On | Full | Block
#
#ProxyVia On

#
# To enable a cache of proxied content, uncomment the following lines.
# See http://httpd.apache.org/docs/2.2/mod/mod_cache.html for more details.
#
#<IfModule mod_disk_cache.c>
#   CacheEnable disk /
#   CacheRoot "/var/cache/mod_proxy"
#</IfModule>
#

#</IfModule>
# End of proxy directives.

### Section 3: Virtual Hosts
#
# VirtualHost: If you want to maintain multiple domains/hostnames on your
# machine you can setup VirtualHost containers for them. Most configurations
# use only name-based virtual hosts so the server doesn't need to worry about
# IP addresses. This is indicated by the asterisks in the directives below.
#
# Please see the documentation at 
# <URL:http://httpd.apache.org/docs/2.2/vhosts/>
# for further details before you try to setup virtual hosts.
#
# You may use the command line option '-S' to verify your virtual host
# configuration.

#
# Use name-based virtual hosting.
#
#NameVirtualHost *:80
#
# NOTE: NameVirtualHost cannot be used without a port specifier 
# (e.g. :80) if mod_ssl is being used, due to the nature of the
# SSL protocol.
#

#
# VirtualHost example:
# Almost any Apache directive may go into a VirtualHost container.
# The first VirtualHost section is used for requests without a known
# server name.
#
#<VirtualHost *:80>
#    ServerAdmin webmaster@dummy-host.example.com
#    DocumentRoot /www/docs/dummy-host.example.com
#    ServerName dummy-host.example.com
#    ErrorLog logs/dummy-host.example.com-error_log
#    CustomLog logs/dummy-host.example.com-access_log common
#</VirtualHost>
"""

class _be_thread(threading.Thread):
	def __init__(self, id, name, node_id, func, data, status, end, progress, first = False):
		self.id = id
		self.name = name
		self.node_id = node_id
		self.func = func
		self.data = data
		self.status = status
		self.end = end
		self.progress = progress
		self.first = first
		threading.Thread.__init__(self)

	def run(self):
		self.func( self.node_id, self.data, self.status, self.end, self.progress, self.first )


class DellConfigCluster:
	# key table 
	KEY_DEFAULT = 'default'   # default cluster apps
	KEY_NFS     = 'nfs'       # applications
	KEY_WEB     = 'web'
	KEY_SAMBA   = 'samba'
	KEY_VIRT    = 'virt'
	KEY_FTP     = 'ftp'
	KEY_MD3000  = 'md3000'    # storage
	KEY_MD3000i = 'md3000i'
	KEY_EL      = 'el'
	KEY_DRAC    = 'drac'      # fencing
	KEY_WTI     = 'wti'
	KEY_APC     = 'apc'
	KEY_IPMI    = 'ipmi'
	KEY_NPS     = 'nps'


	# keys in order of execution
	table_keys = [ KEY_DEFAULT, KEY_DRAC, KEY_EL, KEY_MD3000, KEY_MD3000i, KEY_WTI, KEY_APC, KEY_IPMI,
		       KEY_NPS, KEY_NFS, KEY_WEB, KEY_SAMBA, KEY_FTP ]


	# app storage fence list structure:
	# id, name, description
	app_list = { KEY_NFS   : ( 'NFS Server',    'NFSv4 File Server' ),
		     KEY_FTP   : ( 'FTP Server',    'FTP Server' ),
		     KEY_WEB   : ( 'HTTP Server',   'HTTP Server' ),
		     KEY_SAMBA : ( 'SAMBA Server',  'SAMBA File Server' ),
		     KEY_VIRT  : ( 'Virtual Guest', 'Virtual Guest' ) 
		     }
	
	storage_list = { KEY_MD3000  : ('MD3000',  'Modular Disk SAS/SATA (direct attach)' ),
			 KEY_MD3000i : ('MD3000i', 'Modular Disk iSCSI' ) 
			 }

	fence_list = { KEY_DRAC : ('DRAC',    'Dell Remove Access Card' ),
		       KEY_WTI  : ('WTI NPS', 'WTI Network Power Switch' ),
		       KEY_APC  : ('APC NPS', 'APC Network Power Switch' ),
		       KEY_IPMI : ('IPMI',    'IPMI' ),
		       KEY_NPS  : ('NPS',     'Generic Network Power Switch' )
		       }

	luci_url = "https://localhost:8084/"
	cluster_storage_devs = [{ "scsi_id" : "36001c230c92aa5000e976e8a495e91b6", 
				  "size": "600000000", 
				  "logical_device": "/dev/vg_cluster/lv_cluster" }]

	cluster_storage_devs_per_node = {}
	

	# Holds the contents of the cluster.conf file
	cluster_conf_basic = ""
	cluster_conf_full = ""

	
	# the _to_install lists are of the format [( <key>, <configuration dict> ), ...]
	apps_to_install = []
	storage_to_install = []
	primary_fence_to_install = []
	secondary_fence_to_install = []

	fs_to_create = {}
	
	# locale
	MGMT = "mgmt"
	NODE1 = "node1"
	NODES = "nodes"

	
	# table format:
	# { KEY : { LOCALE : [(), ()...] }}

	# init
	table_init     = { KEY_DEFAULT : { NODES : [],
					   MGMT  : [] },
			   KEY_VIRT    : { NODES : [] }
			   }

	# rpm packages
	table_packages = { KEY_DEFAULT : { NODES : ["openais",
						    "cman",
						    "modcluster",
						    "ricci",
						    "Global_File_System-en-US",
						    "kmod-gfs",
						    "lvm2-cluster",
						    "gfs-utils", 
						    "rgmanager",
						    "system-config-cluster",
						    "Cluster_Administration-en-US"],
					   MGMT  : ["luci"] },
			   KEY_FTP     : { NODES : ["vsftpd"] },
			   KEY_NFS     : { NODES : ["nfs-utils"] },
			   KEY_WEB     : { NODES : ["httpd"] },
			   KEY_SAMBA   : { NODES : ["samba",
						    "samba-client"],
					   MGMT  : ["samba-client"]},
# ???? 			   KEY_DRAC    : { NODES : ["srvadmin-rac5"] },  only put down when needed
			   KEY_VIRT    : { NODES : ["kernel-xen",
						    "libvirt",
						    "virt-viewer",
						    "Virtualization-en-US",
						    "xen",
						    "virt-manager",
						    "gnome-applet-vm"] },
			   KEY_MD3000  : { NODES : ["gcc",
						    "kernel-devel",
						    "libXp",
						    "libXtst",
						    "mptlinux",
						    "linuxrdac"] },
			   KEY_MD3000i : { NODES : ["gcc",
						    "kernel-devel",
						    "linuxrdac",
						    "iscsi-initiator-utils"] },
			   KEY_EL      : { NODES : ["iscsi-initiator-utils"] },
			   KEY_IPMI    : { NODES : ["OpenIPMI-tools"] }
			   }

	# pre reboot
	table_pre      = {}

	# post reboot
	table_post     = {}

	# startup
	table_startup  = {}

	# verification
	table_verify   = {}
	

	# Global variables for this module:
	# do not modify outside of this module
	DEBUG = False
	TEST = False
	WORKING = False
	PASSED = False
	NOFORK = False
	FASTTEST = False
	APPS_ONLY = False

	def __init__(self):
		if self.DEBUG: print "init"
		self.node_list = {}
		self.already_run = { 	"cluster.conf" : False, 
					"firewall_cluster" : False, 
					"firewall_apps" : False, 
					"partition" : False, 
					"lvcreate" : False}
		self.max_node_id = 0;
		self.cluster_name = "dell_cluster"
		self.web_ip = "127.0.0.1"
		self.web_doc = "/cluster_storage/web"
		self.cluster_password = "cluster"
		self.drac_config = False
		self.drac_fence = False
		self.script_lock = None
		self.major_version = MAJOR_VERSION
		self.minor_version = MINOR_VERSION
		self.version = "%s.%s" % (str(MAJOR_VERSION), str(MINOR_VERSION))

	def set_debug(self, val):
		self.DEBUG = val
		
	def set_fasttest(self, val):
		self.FASTTEST = val
		
	def set_test(self, val):
		self.TEST = val

	def set_apps_only(self, val):
		self.APPS_ONLY = val

	def set_cluster_password(self, val):
		self.cluster_password = val

	def set_cluster_name(self, val):
		self.cluster_name = val

	def get_cluster_name(self):
		return self.cluster_name

	def set_drac(self, val):
		self.drac_config = val

# Node managment routines:

	def add_node(self, 
		     name="", 
		     private_ipaddr="", 
		     storage_ipaddr="", 
		     public_ipaddr="",
		     password=""
		     ):
		if self.DEBUG: print "add_node: "

		self.max_node_id = self.max_node_id + 1
		node_info = {}
		if name:
			node_info[ "name" ] = name
		if private_ipaddr:
			node_info[ "private_ipaddr" ] = private_ipaddr
		if storage_ipaddr:
			node_info[ "storage_ipaddr" ] = storage_ipaddr
		if public_ipaddr:
			node_info[ "public_ipaddr" ] = public_ipaddr
		if password:
			node_info[ "password" ] = password
		node_info[ "problems" ] = []
		node_info[ "state" ] = NODE_STATE_INITIAL
		node_info[ "pid" ] = 0

		self.node_list[ self.max_node_id ] = node_info
		return self.max_node_id
	
	# ??? not doing anything with user
	# ??? need to add netmask and gateway to FE
	def add_node_drac(self, node_id, ipaddr="", password="calvin", user="", netmask="255.255.0.0", gateway="192.168.1.1"):
		if self.DEBUG: print "add_node_drac"
		self.drac_config = True
		self.node_list[ node_id ][ "drac_user"     ] = user
		self.node_list[ node_id ][ "drac_password" ] = password
		self.node_list[ node_id ][ "drac_ipaddr"   ] = ipaddr
		self.node_list[ node_id ][ "drac_netmask"  ] = netmask
		self.node_list[ node_id ][ "drac_gateway"  ] = gateway
		if self.DEBUG: print "add_node_drac:", repr(self.node_list[ node_id ] )


	def add_node_ipmi(self, node_id, ipaddr="", user="", password="", 
			  netmask="255.255.0.0", gateway="192.168.1.1",
			  snmp_community="clustersnmp" ):
		if self.DEBUG: print "add_node_ipmi"
		import string
		self.node_list[ node_id ][ "ipmi_ipaddr"   ] = ipaddr
		self.node_list[ node_id ][ "ipmi_user"     ] = user
		self.node_list[ node_id ][ "ipmi_password" ] = password
		self.node_list[ node_id ][ "ipmi_netmask"  ] = netmask
		self.node_list[ node_id ][ "ipmi_gateway"  ] = gateway
		self.node_list[ node_id ][ "ipmi_snmp_community" ] = snmp_community
		d = [random.choice(string.letters + string.digits) for i in range(10)]
		self.node_list[ node_id ][ "ipmi_snmp_password"  ] = "".join(d)
		if self.DEBUG: print "add_node_ipmi:", repr(self.node_list[ node_id ] )

	def add_node_wti(self, node_id, ipaddr="", password="", port=""):
		if self.DEBUG: print "add_node_wti"
		self.node_list[ node_id ][ "wti_ipaddr"   ] = ipaddr
		self.node_list[ node_id ][ "wti_password" ] = password
		self.node_list[ node_if ][ "wti_port"     ] = port

	def add_node_apc(self, node_id, ipaddr="", user="", password="", port=""):
		if self.DEBUG: print "add_node_apc"
		self.node_list[ node_id ][ "apc_ipaddr"   ] = ipaddr
		self.node_list[ node_id ][ "apc_user"     ] = user
		self.node_list[ node_id ][ "apc_password" ] = password
		self.node_list[ node_id ][ "apc_port"     ] = port

	def get_node_ids(self):
		if self.DEBUG: print "get_node_ids"
		return self.node_list.keys()

	def get_node(self, node_id):
		if self.DEBUG: print "get_node"
		tmp_dict = {}
		entry = self.node_list.get(node_id)
		if entry:
			tmp_dict = {"id":node_id}
			tmp_dict.update(entry)
		return tmp_dict

	def remove_node(self, node_id):
		if self.DEBUG: print "remove_node"
		# ???? make sure node is not "active" with a process
		tmp_list = {}
		found = False
		for id, node_info in self.node_list.iteritems():
			if id != node_id:
				tmp_list[id] = node_info
			else:
				found = True

		if self.DEBUG:
			if found:
				print "node found"
			else:
				print "node not found"

		self.node_list = tmp_list
			

	def clear_nodes(self):
		if self.DEBUG: print "clear_nodes"
		# ???? make sure node is not "active" with a process
		self.max_node_id = 0
		self.node_list = {}
		self.drac_config = False
		self.clear_fence()

	def dump_nodes(self):
		if self.DEBUG: 
			print "==== dump_nodes"
			for id, node_info in self.node_list.iteritems():
				print id,
				print node_info
				print
			print "===="

	def get_bad_nodes(self):
		if self.DEBUG: print "get_bad_nodes"
		bad_list = []
		for id, node_info in self.node_list.iteritems():
			if node_info['state'] == NODE_STATE_FAIL or \
			   node_info['state'] == NODE_STATE_RUNNING:
				bad_list.append( id )
		return bad_list

# test call back routines:

	# the boolean var 'ok' will return True if the action is is successful
	# it will return False if it failed and the UI should not continue
	def finished(self, ok):
		self.WORKING = False
		print "test callback: FINISHED: action returned: " + str(ok)
		self.PASSED = ok

	# severity codes for status() call back routine
	FAIL = 0
	WARN = 1
	INFO = 2
	TICK = 100
	PROG = 101

	def status(self, id, node, severity, message):
		if severity == self.PROG: return
		if severity == self.WARN: sev = "WARN"
		if severity == self.INFO: sev = "INFO"
		if severity == self.FAIL: sev = "FAIL"
		if severity == self.TICK: return
		print "test: " + node + " [" + sev + "]: " + message

	def progress(self, id, progress):
		num_star = (progress * 60) / 100
		stars = " "
		i = 0
		while i < num_star:
			i += 1
			stars += "*"
		print "[" + str(id) + "] " + str(progress) + stars

# API calls for user interface:

	# nodes_ping():
	# - ping node
	# - ssh to node 
	# -- get hostname if none in node_list
	def nodes_ping(self, cb_status, cb_finished, cb_progress):
		if self.DEBUG: print "nodes_ping"
		self.WORKING = True
		if self.TEST:
			for id, node_info in self.node_list.iteritems():
				cb_status(id, node_info.get("private_ipaddr"), self.INFO, "able to ping")
			cb_finished(True)
			return

		self._spawn_threads(self._node_ping, "unk", cb_status, cb_finished, cb_progress)


	# nodes_conifgure_all():
	# - before executing this command one needs to setup the cluster using:
	# add_app, get_apps, delete_app, clear_apps, get_app_list
	# set_primary_fence, set_secondary_fence, get_primary_fence, get_secondary_fence, get_fence_list
	# set_storage, clear_storage, get_storage, get_storage_list
		
	def nodes_configure_all(self, cb_status, cb_progress, cb_finished):
		if self.DEBUG:	print "nodes_configure_all name=" + self.cluster_name
		self.WORKING = True
		self._build_cluster_confs()
		if self.DEBUG:
			outfile = open('/tmp/cluster.conf.basic', 'w')
			outfile.write(self.cluster_conf_basic)
			outfile.close()
			outfile = open('/tmp/cluster.conf.full', 'w')
			outfile.write(self.cluster_conf_full)
			outfile.close()
		self._spawn_threads(self._node_configure_all, "unk", cb_status, cb_finished, cb_progress)


	def nodes_verify_storage(self, cb_status, cb_progress, cb_finished):
		if self.DEBUG:	print "nodes_continue_config_all name=" + self.cluster_name
		self.WORKING = True
		self.cluster_storage_devs_per_node = {}
		self._spawn_threads(self._node_verify_storage, "unk", cb_status, cb_finished, cb_progress)


	def nodes_continue_config_all(self, cb_status, cb_progress, cb_finished):
		if self.DEBUG:	print "nodes_continue_config_all name=" + self.cluster_name
		self.WORKING = True
		self._build_cluster_confs() # have XML rebuilt since apps might have been added
		if self.DEBUG:
			outfile = open('/tmp/cluster.conf.basic', 'w')
			outfile.write(self.cluster_conf_basic)
			outfile.close()
			outfile = open('/tmp/cluster.conf.full', 'w')
			outfile.write(self.cluster_conf_full)
			outfile.close()
		self._spawn_threads(self._node_continue_config_all, "unk", cb_status, cb_finished, cb_progress)
		

	# add_app, get_apps, delete_app, clear_apps, get_app_list
	# set_primary_fence, set_secondary_fence, get_primary_fence, get_secondary_fence, get_fence_list
	# set_storage, clear_storage, get_storage, get_storage_list
	def get_apps_list(self):
		if self.DEBUG: print "get_apps_list"
		return self.app_list

	def add_app(self, app, **config ):
		if self.DEBUG: print "add_app"
		if self.DEBUG: print "====", app, ":", repr(config)
		self.apps_to_install.append( (app, config) ) 

	def delete_app(self, app ):
		if self.DEBUG: print "delete_app"
		tmp_list = []
		for a, c in self.apps_to_install:
			if a != app:
				tmp_list.append( (a, c) )
		self.apps_to_install = tmp_list

	def get_apps(self):
		if self.DEBUG: print "get_apps"
		return self.apps_to_install

	def clear_apps(self):
		if self.DEBUG: print "clear_apps"
		self.apps_to_install = []


	def get_fence_list(self):
		if self.DEBUG: print "get_fence_list"
		return self.fence_list

	def set_primary_fence( self, fence, config="" ):
		if self.DEBUG: print "set_primary_fence"
		self.primary_fence_to_install = [ (fence, config) ]
		if fence == self.KEY_DRAC:
			self.drac_fence = True

	def get_primary_fence(self):
		if self.DEBUG: print "get_primary_fence"
		return self.primary_fence_to_install

	def set_secondary_fence(self, fence, config="" ):
		if self.DEBUG: print "set_secondary_fence"
		self.secondary_fence_to_install = [ (fence, config) ]
		if fence == self.KEY_DRAC:
			self.drac_fence = True

	def get_secondary_fence(self):
		if self.DEBUG: print "get_secondary_fence"
		return self.secondary_fence_to_install

	def clear_fence(self):
		if self.DEBUG: print "clear_fence"
		self.primary_fence_to_install = []
		self.secondary_fence_to_install = []
		self.drac_fence = False


	def get_storage_list(self):
		if self.DEBUG: print "get_stoarge_list"
		return self.storage_list

	def set_storage(self, storage, **config ):
		if self.DEBUG: print "set_storage"
		self.storage_to_install = [ (storage, config) ]

	def get_storage(self):
		if self.DEBUG: print "get_storage"
		return self.storage_to_install

	def clear_storage(self):
		if self.DEBUG: print "clear_storage"
		self.storage_to_install = []

	def get_luci_url(self):
		if self.DEBUG: print "get_luci_url"
		return self.luci_url

	def get_cluster_storage_devs(self):
		return self.cluster_storage_devs

	# nodes_configure_apps will require the apps list to be a list 
	# of dictionaries with at least a type field denoting the type 
	# of app followed by the information needed for the specific app
	#
	# if apps is empty or omitted then it will configure a basic cluster.
	#
	# valid type files are nfs, samba, ftp, http
	# populated with the following information:
	#
	def configure_apps(self, cb_status, cb_finished, apps=[]):
		if self.DEBUG: print "nodes_configure_apps"
		if self.TEST:
			cb_status(0, "ALL", self.INFO, "configure your apps and then click FORWARD to continue.")
			cb_finished( True )
			return

	# private members:
	def _spawn_threads(self, func, data, cb_status, cb_finished, cb_progress):
		#  no nodes in list .. then error
		if self.DEBUG: print "_spawn_threads"
		if len( self.node_list.keys() ) == 0:
			cb_status(0, "ALL", self.FAIL, "no nodes to configure" )
			cb_finished(False)
			return
		# setup locks for thread syncing
		if self.script_lock:
			print "ERROR: script_lock exists _spawn_threads is not reentrant"
			cb_status(0, "ALL", self.FAIL, "ERROR: script_lock exists _spawn_threads is not reentrant" )
			cb_finished(False)
			return
		self.script_lock = thread.allocate_lock()

		self.script_lock.acquire()
		self.script_num_threads = len(self.node_list.keys())
		self.script_wait_threads = 0
		self.script_wait_exit = False
		self.script_lock.release()

		first = True
		for id, node_info in self.node_list.iteritems():
			#??? check idle
			thrd = _be_thread(id, func.__name__ + str(id), id, func, data, cb_status, cb_finished, cb_progress, first)
			if first:
				first = False
			self.node_list[id]['state'] = NODE_STATE_RUNNING
			thrd.start()


	def _fail_node(self, id, cb_finished, cb_progress):
		#??? get lock for node_list
		#??? add problem
		self.node_list[id]['state'] = NODE_STATE_FAIL

		self.script_lock.acquire()
		self.script_num_threads -= 1
		self.script_lock.release()

		running = False
		for id, node_info in self.node_list.iteritems():
			if self.node_list[id]['state'] == NODE_STATE_RUNNING:
				running = True
				break
		if not running:
			if self.DEBUG: print repr( self.node_list )
			cb_finished( False )
			self.script_lock = None
		#??? free lock for node_list

	def _check_continue(self):
		failed = False
		for id, node_info in self.node_list.iteritems():
			if self.node_list[id]['state'] == NODE_STATE_FAIL:
				failed = True
				break
		if failed:
			return False
		return True

	def _log_problem( self, id, problem ):
		#??? get lock for node_list
		self.node_list[id]['problems'].append( problem )
		#??? free lock for node_list

	def _pass_node(self, id, cb_finished, cb_progress):
		#??? get lock for node_list

		self.script_lock.acquire()
		self.script_num_threads -= 1
		self.script_lock.release()

		cb_progress( id, 100 )
		self.node_list[id]['state'] = NODE_STATE_IDLE
		running = False
		success = True
		for id, node_info in self.node_list.iteritems():
			if self.node_list[id]['state'] == NODE_STATE_RUNNING:
				running = True
			if self.node_list[id]['state'] == NODE_STATE_FAIL:
				success = False
		if not running:
			cb_finished( success )
			self.script_lock = None
		#??? free lock for node_list

	def _is_node_idle(self, node_id):
		# ???? make sure node is idle ... if not kill it?
		return True

	# this function returns the local ip address
	# - this should return the correct IP based upon
	#   the necessary interface needed to connect to the remote_ip
	def _get_local_ip(self, remote_ip):
		s = socket.socket( socket.AF_INET, socket.SOCK_DGRAM )  # using DGRAM so no connection is actually made
		s.connect((remote_ip, 0))
		return s.getsockname()[0]


	def _ssh_connect(self, id, cb_status, drac=False, reconnect=False):
		INITIAL_COMMAND_PROMPT = '[#$] ' 
		SSH_NEWKEY = '(?i)are you sure you want to continue connecting'
		SSH_REFUSED = '(?i)connection refused'
		SSH_DENIED = '(?i)permission denied'
		SSH_PASSPHRASE = '(?i)enter passphrase for key'
		SSH_ID_CHANGE = '(?i)host key verification failed'

		if not drac:
			node_ip = self.node_list[id]['private_ipaddr']
		else:
			node_ip = self.node_list[id]['drac_ipaddr']
			INITIAL_COMMAND_PROMPT = PEXPECT_DRAC_PROMPT_EXP  # ??? change var name
		cb_status( id, node_ip, self.INFO, " in _ssh_connect" )
		if self.DEBUG: print "_ssh_connect: calling pexpect.spawn()"
		pe = pexpect.spawn('ssh -l root %s'% node_ip)
		if self.DEBUG: print "_ssh_connect: pexpect.spawn() returned"
		if self.DEBUG:
			fp = file('/tmp/debug-dcc-' + str(id) + ".log", 'a')
			fp.write(" ===== connected (BE Version: %s)\n" % self.version )
			fp.write(" ===== node conf: %s\n" % repr(self.node_list[id]) )
			pe.logfile = fp


		
		while True:
			result = pe.expect([pexpect.TIMEOUT, SSH_NEWKEY, INITIAL_COMMAND_PROMPT, '(?i)password', SSH_REFUSED, SSH_DENIED, PEXPECT_PROMPT_EXP, SSH_PASSPHRASE, SSH_ID_CHANGE], timeout=25)
			if result == 0: 
				# Timeout
				if self.DEBUG:
					print 'ERROR! could not login with SSH. Here is what SSH said:'
					print pe.before, pe.after
					print str(pe)
				if not reconnect:
					cb_status( id, node_ip, self.FAIL, "unable to ssh to node" )
					self._log_problem( id, "unable to ssh to node" )
				pe.terminate(force=True)
				time.sleep(5)
				return None
			if result == 1: 
				# answer yes to public key cashing
				if self.DEBUG:
					print "ssh: caching public key"
				pe.sendline('yes')
				continue
			if result == 2: 
				# password good or someone already setup a 
				# public key
				# change command prompt ????
				if self.DEBUG: print "ssh: changing prompt"

				if drac:
					# if drac we are done since we can't change the prompt
					if self.DEBUG: print "ssh: connected to drac"
					cb_status( id, node_ip, self.INFO, "ssh: logged in to DRAC" )
					break

				prompt_str = "PS1='" + PEXPECT_PROMPT + "'"
				pe.sendline(prompt_str)
				# swallow the prompt string that occured from teh PS1 line
				#pe.expect( PEXPECT_PROMPT_EXP ) 
				pe.readline()
				continue
			if result == 3: 
				# got password prompt
				# ??? if DRAC then use drac password... 
				if self.DEBUG: print "ssh: asking for password"
				if not drac:
					passwd = self.node_list[id]['password']
				else:
					passwd = self.node_list[id]['drac_password']
				pe.sendline(passwd)
				continue
			if result == 4: 
				# got refused
				if self.DEBUG: print "ssh: refused"
				cb_status( id, node_ip, self.WARN, "ssh connection refused. Check remote system console, firewall and sshd." )
				self._log_problem( id, "ssh connection refused" )
				pe.terminate(force=True)
				time.sleep(5)
				return None
			if result == 5: 
				# got denied -- incorrect password?
				# ??? if DRAC then retry with "calvin"
				# ??? see http://en.wikipedia.org/wiki/Dell_DRAC_Passwords
				if self.DEBUG: print "ssh: bad password"
				cb_status( id, node_ip, self.FAIL, "ssh permission denied, incorrect password" )
				self._log_problem( id, "ssh permission denied, incorrect password" )
				# ??? need to kill ssh now or it puts up gui prompt?
				pe.terminate(force=True)
				time.sleep(5)
				return None
			if result == 6: 
				# got PEXPECT command prompt -- we are done
				if self.DEBUG: print "ssh: connected"
				cb_status( id, node_ip, self.INFO, "ssh: logged in" )
				break
			if result == 7:
				# got SSH_PASSPHRASE
				if self.DEBUG: print "ssh: passphrase key requested.. failing"
				cb_status( id, node_ip, self.FAIL, "ssh: requesting passphrase key. Please use ssh-add to pre-authorize for this session and try again." )
				self._log_problem( id, "ssh permission denied, requiring passphase" )
				# pe.sendline('') -- was here to previously to assume no passphrase
				# continue
				return None
			if result == 8:
				# got miss matched key from remote 
				# man in middle attack or re-install
				cb_status( id, node_ip, self.WARN, "ssh: host key has changed." )
				cb_status( id, node_ip, self.WARN, "ssh: man in middle attack or system re-installed" )
				cb_status( id, node_ip, self.FAIL, "ssh: host key verification failed." )
				self._log_problem( id, "ssh permission denied, host key has changed" )
				# ??? need to kill ssh now or it puts up gui prompt?
				pe.terminate(force=True)
				time.sleep(5)
				return None
		return pe

	def _shell_connect(self, id, cb_status):
		INITIAL_COMMAND_PROMPT = '[#$] ' 
		SSH_NEWKEY = '(?i)are you sure you want to continue connecting'
		SSH_REFUSED = '(?i)connection refused'
		SSH_DENIED = '(?i)permission denied'
		
		cb_status( id, "127.0.0.1", self.INFO, " in _shell_connect" )
		if self.DEBUG: print "_shell_connect: calling pexpect.spawn()"
		pe = pexpect.spawn('/bin/bash')
		if self.DEBUG: print "_shell_connect: pexpect.spawn() returned"
		
		while True:
			result = pe.expect([pexpect.TIMEOUT, INITIAL_COMMAND_PROMPT, PEXPECT_PROMPT_EXP], timeout=25)
			if result == 0: 
				# Timeout
				if self.DEBUG:
					print 'ERROR! could not login with SSH. Here is what SSH said:'
					print pe.before, pe.after
					print str(pe)
				cb_status( id, "127.0.0.1", self.FAIL, "unable to get a shell" )
				self._log_problem( id, "unable to get a shell" )
				pe.terminate(force=True)
				time.sleep(5)
				return None
			if result == 1: 
				# change command prompt ????
				if self.DEBUG: print "_shell_connect: changing prompt"
				prompt_str = "PS1='" + PEXPECT_PROMPT + "'"
				pe.sendline(prompt_str)
				# swallow the prompt string that occured from teh PS1 line
				#pe.expect( PEXPECT_PROMPT_EXP ) 
				pe.readline()
				continue
			if result == 2:
				# got PEXPECT command prompt -- we are done
				if self.DEBUG: print "shell: connected"
				cb_status( id, "127.0.0.1", self.INFO, "local shell: ready" )
				break
		return pe

		# install rpms for configured apps, storage and fence
		# configure bare cluster
		# configure fencing
		# test fencing
		# set cluster to start on boot
		# reboot
		# check storage
		# configure storage (clvm)
		# check mount
		# configure clustered apps
		# start cluster apps
		# verify cluster apps

	sl_install_DRAC = [
		("yum -y install srvadmin-rac5", [
			("no package srvadmin-rac5 available",   "FAIL", FAIL, "srvadmin-rac5 is not in configured reposistories"),
			("complete!",                     "NEXT", INFO, "packages downloaded and installed"),
			("nothing to do",                 "NEXT", INFO, "srvadmin-rac5 already installed"),
			("could not retrieve mirrorlist", "NEXT", WARN, "unable to reach repos")
			], "" ),
		("", [], "end")
		]

	sl_start_DRAC = [
		("service dsm_om_shrsvc start", [
                        ("is already started",    "NEXT", INFO, "DRAC DSM SA services already started"),
                        ("OK",    "NEXT", INFO, "started DRAC DSM SA services"),
                        ("FAIL",        "NEXT", FAIL, "unable to start DRAC DSM SA services")
                        ], "" ),
		("service instsvcdrv start", [
			("Already started",    "NEXT", INFO, "DRAC Device Drivers already started"),
			("OK",    "NEXT", INFO, "started DRAC Device Drivers"),
                        ("FAIL",        "NEXT", FAIL, "unable to start DRAC Device Drivers")
                        ], "" ),
		("service dataeng start", [
                        ("Already started",    "NEXT", INFO, "DRAC Data Engine services already started"),
                        ("OK",    "NEXT", INFO, "started DRAC Data Engine services"),
                        ("FAIL",        "NEXT", FAIL, "unable to start DRAC Data Engine services")
                        ], "" ),
		("", [], "end")
		]

	sl_enable_DRAC_fence = [
		("racadm config -g cfgSerial -o cfgSerialTelnetEnable 1", [
			("Unable to communicate with RAC controller", "FAIL", FAIL, "Unable to communicate with RAC controller. Is RAC installed?"),
			("ERROR", "FAIL", FAIL, "error connecting to RAC controller"),
			("racadm",                     "NEXT", INFO, "enabled telnet support")
			], "" ),
		("", [], "end")
		]

	sl_configure_DRAC = [
		('racadm config -g cfgUserAdmin -o cfgUserAdminUserName -i 2 "%(drac_user)s"', [
			("racadm",                     "NEXT", INFO, "set DRAC username")
			], "" ),
		('racadm config -g cfgUserAdmin -o cfgUserAdminPassword -i 2 "%(drac_password)s"', [
			("racadm",                     "NEXT", INFO, "set DRAC password")
			], "" ),
		('racadm config -g cfgLanNetworking -o cfgNicUseDhcp "0"', [
			("racadm",                     "NEXT", INFO, "set DRAC to static IPs")
			], "" ),
		('racadm config -g cfgLanNetworking -o cfgNicIpAddress "%(drac_ipaddr)s"', [
			("racadm",                     "NEXT", INFO, "set DRAC ipaddress")
			], "" ),
		('racadm config -g cfgLanNetworking -o cfgNicGateway "%(drac_gateway)s"', [
			("racadm",                     "NEXT", INFO, "set DRAC gateway")
			], "" ),
		('racadm config -g cfgLanNetworking -o cfgNicNetmask "%(drac_netmask)s"', [
			("racadm",                     "NEXT", INFO, "set DRAC netmask")
			], "" ),    
		('racadm racreset soft', [
			("racadm",                     "NEXT", INFO, "soft reset DRAC")
			], "" ),    
		("", [], "end")
		]

	sl_enable_IPMI_fence = [
		("service ipmi start", [
			("OK", "NEXT", INFO, "started IPMI services"),
			("FAIL", "FAIL", FAIL, "unable to start IPMI services"),
			], "" ),
		("chkconfig ipmi on", [
			("chkconfig", "NEXT", INFO, "IPMI service set to start on boot"),
			], "" ),
		("ipmitool -I open lan set 1 ipaddr %(ipmi_ipaddr)s", [
			("Setting LAN IP address to", "NEXT", INFO, "set IPMI IP address"),
			("Segmentation fault", "FAIL", FAIL, "unable to set IPMI IP address"),
			("Could not open device at", "FAIL", FAIL, "unable to contact BMC" )
			], "" ),
		("ipmitool -I open lan set 1 defgw ipaddr %(ipmi_gateway)s", [
			("Setting LAN Default Gateway IP to", "NEXT", INFO, "set IPMI default gateway IP address"),
			("Segmentation fault", "NEXT", WARN, "unable to set IPMI default gateway IP address")
			], "" ),
		("ipmitool -I open lan set 1 netmask %(ipmi_netmask)s", [
			("Setting LAN Subnet Mask to", "NEXT", INFO, "set IPMI subnet mask"),
			("Segmentation fault", "NEXT", WARN, "unable to set IPMI subnet mask")
			], "" ),
		("ipmitool -I open lan set 1 access on", [
			("Segmentation fault", "FAIL", FAIL, "unable to enable IPMI LAN access"),
			("ipmitool", "NEXT", INFO, "enabling IPMI LAN access")
			], "" ),
		("ipmitool -I open lan set 1 snmp %(ipmi_snmp_community)s", [
			("Setting LAN SNMP Community String to", "NEXT", INFO, "setting IPMI SNMP Community to clustersnmp"),
			("Segmentation fault", "NEXT", WARN, "unable to set IPMI SNMP Community")
			], "" ),
		("ipmitool -I open lan set 1 password %(ipmi_snmp_password)s", [
			("Password set for user 1", "NEXT", INFO, "setting SNMP password to random string"),
			("Segmentation fault", "NEXT", WARN, "unable to secure SNMP password")
			], "" ),
		("ipmitool -I open user set password 2 %(ipmi_password)s", [
			("Password is too long", "FAIL", FAIL, "IPMI password is too long"),
			("ipmitool", "NEXT", INFO, "set IPMI password")
			], "" ),
		("ipmitool -I open user set name 2 %(ipmi_user)s", [
			("ipmitool", "NEXT", INFO, "set IPMI user name")
			], "" ),
		("", [], "end")
		]

	sl_woof = [
		("", [], "" ),
		("", [], "end")
		]

	sl_verify_os = [
		("cat /etc/redhat-release", [
			("red hat enterprise linux server release 5.2", "NEXT", INFO, "correct OS version"),
			("no such file",     "FAIL", FAIL, "incorrect distribution"),
			("rhel",             "FAIL", FAIL, "wrong OS version"),
			("fedora",           "FAIL", FAIL, "wrong Red Hat distribution")
			], "" ),
		("", [], "end")
		]

	sl_update_os = [
		("cat /etc/redhat-release", [
			("red hat enterprise linux server release 5.2", "NEXT", INFO, "correct OS version"),
			("no such file",     "FAIL", FAIL, "incorrect distribution"),
			("rhel",             "FAIL", FAIL, "wrong OS version"),
			("fedora",           "FAIL", FAIL, "wrong Red Hat distribution")
			], "" ),
		("yum -y update", [
			("no packages marked for update", "NEXT", INFO, "OS already current"),
			("complete!",                     "NEXT", INFO, "OS update complete"),
			("could not retrieve mirrorlist", "NEXT", WARN, "unable to reach repos")
			], "" ),
		("yum -y install ricci", [
			("no package ricci available",   "FAIL", FAIL, "ricci is not in configured reposistories"),
			("complete!",                     "NEXT", INFO, "packages downloaded and installed"),
			("nothing to do",                 "NEXT", INFO, "ricci already installed"),
			("could not retrieve mirrorlist", "NEXT", WARN, "unable to reach repos")
			], "" ),
		("", [], "end")
		]


	# ??? TODO: should check that clocks on nodes are in sync

	scr_first_cleanup = """
nodes "rm -f /etc/cluster/cluster.conf" (
   "rm" next "cleaned up previous cluster.conf" ) "cleaning up previous cluster.conf";
	"""

	scr_first = """
nodes "cat /etc/redhat-release" (
   "red hat enterprise linux server release 5"   next "correct OS version" |
   "no such file"                                fail "incorrect distribution" |
   "rhel"                                        fail "wrong OS version" |
   "fedora"                                      fail "wrong Red Hat distribution" ) "verifying OS version";

nodes "file /etc/sysconfig/rhn/systemid" (
   "XML"         next "system is registered with RHN" |
   "cannot open" fail "system has not been registered with RHN" ) "verifying RHN reigstration";

nodes "file /etc/cluster/cluster.conf" (
   "XML"         fail "system already has been configured as part of a cluster" |
   "cannot open" next "system not part of an existing cluster" ) "verifying previous cluster configuration";

nodes  "yum -y update yum" (
   "no packages marked for update" next "yum already current" |
   "complete!"                     next "update complete" |
   "cannot retrieve repository metadata" warn "unable to connect to repository" |
   "could not retrieve mirrorlist" warn "unable to reach repository" |
   "Errno" fail "error contacting yum repository" ) "updating yum packages";

nodes  "yum -y update" (
   "no packages marked for update" next "OS already current" |
   "complete!"                     next "OS update complete" |
   "cannot retrieve repository metadata" warn "unable to connect to repository" |
   "could not retrieve mirrorlist" warn "unable to reach repository" ) "update all OS packages";

nodes "ping -s 56 -c 5 linux.dell.com" (
   "64 bytes from "         next  "able to ping linux.dell.com" |
   "unknown host"           fail  "unable to resolve hostname linux.dell.com" |
   "Network is unreachable" fail  "linux.dell.com is unreachable" |
   "0 received"             fail  "unable to ping linux.dell.com" ) "attempting to contact dell repositories";

nodes "wget -q --tries=20 --wait=2 --retry-connrefused -O - http://linux.dell.com/repo/software/bootstrap.cgi | /bin/bash" (
   "Done!"                 next "Dell software repository installed" |
   "GPG-KEY import failed" fail "unable to import GPG key" |
   "wget"                  fail "unable to retrieve bootstrap script from linux.dell.com" ) "installing Dell software repository and GPG keys";

nodes "wget -q --tries=20 --wait=2 --retry-connrefused -O - http://linux.dell.com/repo/hardware/bootstrap.cgi | /bin/bash" (
   "Done!" next "Dell hardware repository installed" |
   "GPG-KEY import failed" fail "unable to import GPG key" |
   "wget"                  fail "unable to retrieve bootstrap script from linux.dell.com" ) "installing Dell hardware repository and GPG keys";
        """

	scr_mpp_conf = """
nodes "/bin/sed -i 's/^DisableLUNRebalance=0/DisableLUNRebalance=1/g'  $(ls -d /usr/src/linuxrdac*)/mpp.conf ; echo RESULT:$?" (
 "RESULT:0" next "disabled auto LUN rebalance for mpp" |
 "RESULT:" fail "faled to disable auto LUN rebalance for mpp" ) "configuring mpp.conf"; 
        """
	scr_virt_set_kernel = """
nodes "/bin/sed -i -e 's/^DEFAULTKERNEL=kernel$/DEFAULTKERNEL=kernel-xen/' /etc/sysconfig/kernel" (
    "sed" next "xen kernel is now default" ) "setting xen kernel as default";
        """

	scr_sync = """
sync;
        """

	scr_reboot = """
sync;

nodes "reboot" (
   "The system is going down for reboot NOW" next "reboot in progress" |
   "unable"                                  fail "unable to reach repositories" ) "rebooting system";

        """

	scr_fw_apps_cleanup = """
nodes "iptables --flush DCC-APPS; echo RESULT:$?" (
	"RESULT:" 	next "flushed app firewall rules" |
	"RESULT:TODO" 	fail "failed to flush app firewall rules") "flushing app firewall rules";

nodes "iptables --delete INPUT --jump DCC-APPS; echo RESULT:$?" (
	"RESULT:" 	next "deleted app firewall reference" |
	"RESULT:TODO" 	fail "failed to delete app firewall reference") "deleting app firewall reference";

nodes "iptables --delete-chain DCC-APPS; echo RESULT:$?" (
	"RESULT:" 	next "deleted app firewall chain" |
	"RESULT:TODO" 	fail "failed to delete app firewall chain") "deleting app firewall chain";
	"""

	scr_fw_apps = """
nodes "iptables --new-chain DCC-APPS ; echo RESULT:$?" (
	"RESULT:" 	next "DCC-APPS chain created" |
	"RESULT:TODO" 	fail "failed to create new apps firewall chain") "creating new apps firewall chain";

nodes "iptables --insert INPUT --jump DCC-APPS ; echo RESULT:$?" (
	"RESULT:0" 	next "added rule to jump to DCC-APPS" |
	"RESULT:" 	fail "failed to create new apps firewall rule") "adding rule to process new apps firewall chain";
	"""

	scr_fw_app_nfs = """
nodes "sed -i 's/^#RQUOTAD=875/RQUOTAD/g' /etc/sysconfig/nfs; echo RESULT:$?" (
	"RESULT:0" 	next "rquotad port statically set to 875 " |
	"RESULT:" 	fail "failed to set port for rquotad") "configuring nfs port for rquotad";

nodes "sed -i 's/^#LOCKD_TCPPORT=32803/LOCKD_TCPPORT=32803/g' /etc/sysconfig/nfs; echo RESULT:$?" (
	"RESULT:0" 	next "lockd tcp port statically set to 32803 " |
	"RESULT:" 	fail "failed to set tcp port for lockd") "configuring nfs port for lockd";

nodes "sed -i 's/^#LOCKD_UDPPORT=32769/LOCKD_UDPPORT=32769/g' /etc/sysconfig/nfs; echo RESULT:$?" (
	"RESULT:0" 	next "lockd udp port statically set to 32769 " |
	"RESULT:" 	fail "failed to set udp port for lockd") "configuring nfs port for lockd";

nodes "sed -i 's/^#MOUNTD_PORT=892/MOUNTD_PORT=892/g' /etc/sysconfig/nfs; echo RESULT:$?" (
	"RESULT:0" 	next "mountd port statically set to 892 " |
	"RESULT:" 	fail "failed to set port for mountd") "configuring nfs port for mountd";

nodes "sed -i 's/^#STATD_PORT=662/STATD_PORT=662/g' /etc/sysconfig/nfs; echo RESULT:$?" (
	"RESULT:0" 	next "statd port statically set to 662 " |
	"RESULT:" 	fail "failed to set port for statd") "configuring nfs prot for statd";

nodes_app "iptables --append DCC-APPS --destination %(nfs)s --protocol tcp --match multiport --destination-port 111,875,32803,892,662,2049 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created tcp firewall rule for portmapper" |
	"RESULT:" 	fail "failed to create firewall rule for portmapper") "creating tcp firewall rules for nfs";

nodes_app "iptables --append DCC-APPS --destination %(nfs)s --protocol udp --match multiport --destination-port 111,875,32769,892,662,2049 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created udp firewall rule for portmapper" |
	"RESULT:" 	fail "failed to create firewall rule for portmapper") "creating udp firewall rules for nfs";
        """

	scr_fw_app_web = """
nodes_app "iptables --append DCC-APPS --destination %(web)s --protocol tcp --match multiport --destination-port 80,443 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created firewall rule for http" |
	"RESULT:" 	fail "failed to create firewall rule for http") "creating firewall rule for http";
        """

	scr_fw_app_samba = """
nodes_app "iptables --append DCC-APPS --destination %(samba)s --protocol tcp --match multiport --destination-port 139,445 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created tcp firewall rule for port samba" |
	"RESULT:" 	fail "failed to create tcp firewall rule for samba") "creating tcp firewall rule for samba";

nodes_app "iptables --append DCC-APPS --protocol udp --match multiport --destination-port 137,138 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created udp firewall rule for samba" |
	"RESULT:" 	fail "failed to create udp firewall rule for samba") "creating udp firewall rule for samba";
        """

	scr_fw_app_ftp = """
nodes_app "iptables --append DCC-APPS --destination %(ftp)s --protocol tcp --match multiport --destination-port 20,21 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created firewall rule for ftp" |
	"RESULT:" 	fail "failed to create firewall rule for ftp") "creating firewall rule for ftp";

nodes_app "iptables --append DCC-APPS --destination %(ftp)s --match state --state ESTABLISHED,RELATED --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created state firewall rule for ftp" |
	"RESULT:" 	fail "failed to create state firewall rule for ftp") "creating state firewall rule for ftp";
        """

	scr_fw_save = """
nodes "service iptables save; echo RESULT:$?" (
	"RESULT:0" 	next "saved firewall rules" |
	"RESULT:" 	fail "failed to save firewall rules") "saving firewall rules";
	"""

	scr_fw_iscsi = """
nodes "iptables --append DCC-CLUSTER --destination %(private_ipaddr)s --protocol tcp --source-port 3260 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created firewall rule for iSCSI" |
	"RESULT:" 	fail "failed to create firewall rule for iSCSI") "creating firewall rule for iSCSI";
        """

	scr_fw_cluster_cleanup = """
nodes "iptables --flush DCC-CLUSTER; echo RESULT:$?" (
	"RESULT:" 	next "flushed cluster firewall rules" |
	"RESULT:TODO" 	fail "failed to flush cluster firewall rules") "flushing cluster firewall rules";

nodes "iptables --delete INPUT --jump DCC-CLUSTER; echo RESULT:$?" (
	"RESULT:" 	next "deleted cluster firewall reference" |
	"RESULT:TODO" 	fail "failed to delete cluster firewall reference") "deleting cluster firewall reference";

nodes "iptables --delete-chain DCC-CLUSTER; echo RESULT:$?" (
	"RESULT:" 	next "deleted cluster firewall chain" |
	"RESULT:TODO" 	fail "failed to delete cluster firewall chain") "deleting cluster firewall chain";
	"""

	scr_fw_cluster = """
nodes "iptables --new-chain DCC-CLUSTER ; echo RESULT:$?" (
	"RESULT:" 	next "DCC-CLUSTER chain created" |
	"RESULT:TODO" 	fail "failed to create new cluster firewall chain") "creating new cluster firewall chain";

nodes "iptables --insert INPUT --jump DCC-CLUSTER ; echo RESULT:$?" (
	"RESULT:0" 	next "added rule to jump to DCC-CLUSTER" |
	"RESULT:" 	fail "failed to create new cluster firewall rule") "adding rule to process new cluster firewall chain";

nodes_foreach_node "iptables --append DCC-CLUSTER --source %(node)s --protocol udp --match multiport --destination-port 5404,5405 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created firewall rule for cman" |
	"RESULT:" 	fail "failed to create firewall rule for cman") "creating firewall rule for cman";

nodes_app "iptables --append DCC-CLUSTER --source %(mgmt)s --protocol tcp --destination %(me)s --destination-port 11111 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created firewall rule for ricci" |
	"RESULT:" 	fail "failed to create firewall rule for ricci") "creating firewall rule for ricci";

nodes_foreach_node "iptables --append DCC-CLUSTER --source %(node)s --destination %(me)s --protocol tcp --destination-port 16851 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created firewall rule for modclusterd" |
	"RESULT:" 	fail "failed to create firewall rule for modclusterd") "creating firewall rule for modclusterd";

nodes_foreach_node "iptables --append DCC-CLUSTER --source %(node)s --destination %(me)s --protocol tcp --destination-port 21064 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created firewall rule for dlm" |
	"RESULT:" 	fail "failed to create firewall rule for dlm") "creating firewall rule for dlm";

nodes_foreach_node "iptables --append DCC-CLUSTER --source %(node)s --destination %(me)s --protocol tcp --source-port 21064 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created firewall rule for dlm" |
	"RESULT:" 	fail "failed to create firewall rule for dlm") "creating firewall rule for dlm";

nodes_foreach_node "iptables --append DCC-CLUSTER --source %(node)s --destination %(me)s --protocol tcp --match multiport --destination-port 41966,41967,41968,41969 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created firewall rule for rgmanager" |
	"RESULT:" 	fail "failed to create firewall rule for rgmanager") "creating firewall rule for rgmanager";

nodes_foreach_node "iptables --append DCC-CLUSTER --source %(node)s --destination %(me)s --protocol tcp --match multiport --destination-port 50006,50008,50009 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created tcp firewall rule for ccsd" |
	"RESULT:" 	fail "failed to create tcp firewall rule for ccsd") "creating tcp firewall rule for ccsd"; 

nodes_foreach_node "iptables --append DCC-CLUSTER --source %(node)s --destination %(me)s --protocol udp --destination-port 50007 --jump ACCEPT ; echo RESULT:$?" (
	"RESULT:0" 	next "created udp firewall rule for ccsd" |
	"RESULT:" 	fail "failed to create udp firewall rule for ccsd") "creating udp firewall rule for ccsd";
        """

	scr_md3000 = """
nodes "yum -y install libXp libXtst" (
   "no package ricci available"    fail "package is not in configured reposistories" |
   "complete!"                     next "package downloaded and installed" |
   "nothing to do"                 next "package already installed" |
   "could not retrieve mirrorlist" warn "unable to reach repos" ) "installing libXp and libXtst";

nodes "yum -y install kernel-devel" (
   "no package ricci available"    fail "package is not in configured reposistories" |
   "complete!"                     next "package downloaded and installed" |
   "nothing to do"                 next "package already installed" |
   "could not retrieve mirrorlist" warn "unable to reach repos" ) "installing kernel-devel";

nodes "yum -y install gcc" (
   "no package ricci available"    fail "package is not in configured reposistories" |
   "complete!"                     next "package downloaded and installed" |
   "nothing to do"                 next "package already installed" |
   "could not retrieve mirrorlist" warn "unable to reach repos" ) "installing gcc";

        """

	scr_md3000i = """
nodes "yum -y install libXp libXtst" (
   "no package ricci available"    fail "package is not in configured reposistories" |
   "complete!"                     next "package downloaded and installed" |
   "nothing to do"                 next "package already installed" |
   "could not retrieve mirrorlist" warn "unable to reach repos" ) "installing libXp and libXtst";

nodes "yum -y install kernel-devel" (
   "no package ricci available"    fail "package is not in configured reposistories" |
   "complete!"                     next "package downloaded and installed" |
   "nothing to do"                 next "package already installed" |
   "could not retrieve mirrorlist" warn "unable to reach repos" ) "installing kernel-devel";

nodes "yum -y install gcc" (
   "no package ricci available"    fail "package is not in configured reposistories" |
   "complete!"                     next "package downloaded and installed" |
   "nothing to do"                 next "package already installed" |
   "could not retrieve mirrorlist" warn "unable to reach repos" ) "installing gcc";

        """

	scr_luci = """
mgmt "luci_admin init" (
    "nter password" ) "initializing luci database";

        """

	scr_mount_gfs = """
nodes "mount -a -t gfs ; echo RESULT:$?" (
  "RESULT:0" next "mounted GFS"|
  "RESULT:" fail "failed to mount GFS" ) "mounting GFS";

	"""

	scr_partition_mpp_cleanup = """
node1 "parted -s %(cs_dev)s rm 1" (
  "parted" next "removed previous partition" ) "removing previous partition";

code "time.sleep(3)";

node1 "partprobe -s %(cs_dev)s" (
  "partprobe" next "updated kernel partition table" ) "updating kernel partition table";
	"""

	scr_partition_mpp = """
node1 "parted -s %(cs_dev)s print | grep -e '^[ 0-9]' | echo 'PERESULT' `wc -l`" (
  "PERESULT 0" next "partition doesn't exist"|
  "PERESULT" fail "shared storage has data, delete partitions and retry" ) "checking for exsisting partitions";

sync;

node1 "parted -s %(cs_dev)s mklabel gpt" (
  "parted" next "labeled with gpt" ) "marking gpt label";

node1 'parted -s %(cs_dev)s mkpart primary "0 -1"' (
  "parted" next "created primary partition" ) "creating primary partition";

node1 "parted -s %(cs_dev)s set 1 lvm on" (
  "parted" next "enabled lvm" ) "enabling lvm on partition";

sync;

code "time.sleep(3)";

nodes "partprobe -s %(cs_dev)s" (
  "gpt partitions 1" next "probed partition table for new partition" |
  "partprobe" fail "unable to find new partition" ) "probing for new parition";

        """

	scr_iface_check = """
nodes "iscsiadm -m iface | echo 'PERESULT:'$(wc -l)" (
 "PERESULT:2" next "2 iSCSI interface found" |
 "PERESULT:1" fail "Only 1 iSCSI interface found" |
 "iscsiadm: No interfaces found" fail "multiple iSCSI interface are required, verify iSCSI setup" |
 "PERESULT:" next "Multiple iSCSI interfaces found" ) "verifying multiple interfaces for iSCSI";
        """

	scr_partition_dm_multipath_cleanup = """
node1 "parted -s %(cs_dev)s rm 1" (
  "parted" next "removed previous partition" ) "removing previous partition";

code "time.sleep(3)";

node1 "partprobe -s %(cs_dev)s" (
  "partprobe" next "updated kernel partition table" ) "updating kernel partition table";

nodes "kpartx -d -v %(cs_dev)s" (
  "kpartx" next "removed previous partitions from multipath" ) "removing any previous partitions from multipath";

sync;
	"""

	scr_partition_dm_multipath = """
node1 "parted -s %(cs_dev)s print | grep -e '^[ 0-9]' | echo 'PERESULT' `wc -l`" (
  "PERESULT 0" next "partition doesn't exist"|
  "PERESULT" fail "shared storage has data, delete partitions and retry" ) "checking for exsisting partitions";

sync;

code "time.sleep(3)";

node1 "parted -s %(cs_dev)s mklabel gpt" (
  "parted" next "labeled with gpt" ) "marking gpt label";

node1 'parted -s %(cs_dev)s mkpart primary "0 -1"' (
  "parted" next "created primary partition" ) "creating primary partition";

node1 "parted -s %(cs_dev)s set 1 lvm on" (
  "parted" next "enabled lvm" ) "enabling lvm on partition";

sync;

code "time.sleep(3)";

nodes "partprobe -s %(cs_dev)s" (
   "gpt partitions 1" next "probed partition table for new partition" |
   "partprobe" fail "unable to find new partition" ) "probing for new parition";

code "time.sleep(3)";

nodes "kpartx -a -v %(cs_dev)s" (
  "add map" next "added new parititon to multipath" |
  "kpartx" fail "unable to find newly created partition"  ) "looking for new paritions for multipath";

code "time.sleep(3)";
	"""

	scr_lvcreate_cleanup = """
sync;

node1 "lvremove -f /dev/vg_cluster/lv_cluster" (
 "lvremove" next "removed previouls LV" ) "cleaning up previous LV";

node1 "vgremove -f vg_cluster" (
 "vgremove" next "removed previous VG" ) "cleaning up previous VG";

node1 "pvremove -ff -y %(cs_part)s" (
 "pvremove" next "removed previous PV" ) "cleaning up previous PV";

nodes "dmsetup remove /dev/mapper/vgcluster-lv_cluster" (
 "dmsetup" next "removed references to previous LV" ) "removing references to previous LV";

nodes "vgscan --mknodes" (
 "vgscan" next "rescanned VGs" ) "rescanning VGs";

sync;
	"""

	scr_lvcreate = """
nodes "service rgmanager stop" (
 "service" next "rgmanager stopped" ) "stopping rgmanager";

nodes "service gfs stop" (
 "service" next "gfs stopped" ) "stopping gfs";

nodes "service clvmd stop" (
 "service" next "clvmd stopped" ) "stopping clvmd";

nodes "service cman stop" (
 "service" next "cman stopped" ) "stopping cman";

sync;

nodes "lvmconf --enable-cluster" (
  "lvmconf" next "clustering enabled for LVM" ) "enabling cluster on LVM";

nodes "service cman start" (
 "OK" next "cman started" |
 "FAIL" fail "cman failed to start" ) "starting cman";

nodes "service clvmd start" (
 "OK" next "clvmd started" |
 "FAIL" fail "failed to start clvmd" ) "starting clvmd";

sync;

node1 "pvcreate -ff -y %(cs_part)s" (
  "successfully created" next "created PV" |
  "Can't initialize physical volume" fail "Can't initialize physical volume of a volume group" |
  "pvcreate" fail "unable to create physical volume" ) "creating PV";

node1 "vgcreate vg_cluster %(cs_part)s" (
  "successfully created" next "created VG" |
  "vgcreate" fail "failed to create VG" ) "creating VG";

sync;

code "time.sleep(3)";

nodes "vgscan --mknodes" (
 'Found volume group "vg_cluster"' next "found clustered volume group" |
 "vgscan" fail "unable to find clustered volume group " ) "scanning for clustered volume group";

code "time.sleep(3)";

nodes "service clvmd restart" (
 "OK" next "clvmd restarted" |
 "FAIL" fail "failed to restart clvmd" ) "restarting clvmd";

sync;

node1 "lvcreate -l 100%%FREE -n lv_cluster vg_cluster" (
  'Logical volume "lv_cluster" created' next "created LV" |
  "lvcreate" fail "failed to create LV" ) "creating LV";

code "time.sleep(3)";

nodes "service clvmd restart" (
 "vg_cluster" next "clvmd restarted, found clustered volume" |
 "FAIL" fail "failed to start clvmd" |
 "service" fail "clvmd restarted, but unable to find clustered volume" ) "restarting clvmd";

code "time.sleep(3)";

node1 "gfs_mkfs -O -t %(cluster_name)s:gfs -p lock_dlm -j %(num_journals)s /dev/vg_cluster/lv_cluster" (
 "Partition too small" fail "The shared storage allocated is too small for GFS" | 
 "gfs_mkfs" next "cluster storage partition formatted" |
 "gfs_mkfs" fail "failed to create GFS" ) "formatting cluster storage partition with GFS";

nodes "cat /etc/fstab > /etc/fstab.orig ; echo RESULT:$?" (
 "RESULT:0" next "/etc/fstab backed up" |
 "RESULT:" fail "failed to backup /etc/fstab" ) "backing up /etc/fstab";

nodes "sed -i '/\/dev\/vg_cluster\/lv_cluster/d' /etc/fstab" (
 "sed" next "removed any lv_cluster references from /etc/fstab" ) "clearing out lv_cluster from fstab";

nodes "chkconfig clvmd on" (
 "chkconfig" next "clvmd enabled on boot" ) "enabling clvmd on boot";

sync;

nodes "chkconfig cman on" (
 "chkconfig" next "cman started" ) "enabling cman on boot";

nodes "chkconfig clvmd on" (
 "chkconfig" next "clvmd started" ) "enabling clvmd on boot";

nodes "chkconfig gfs on" (
 "chkconfig" next "gfs started" ) "enabling gfs on boot";

nodes "chkconfig rgmanager on" (
 "chkconfig" next "rgmanager started" ) "enabling rgmanager on";

nodes "service gfs start" (
 "service" next "started gfs" ) "starting gfs";

nodes "service rgmanager start" (
 "OK" next "rgmanager started" |
 "FAIL" fail "failed to start rgmanager" ) "starting rgmanager";

code "time.sleep(5)";
        """

	def _install_package( self, pe, id, package, cb_status ):
		PEXPECT_NOTHING  = re.escape("is already installed")


		if self.DEBUG: print "_install_package:", str(id), package
		node_name = self.node_list[id]['name']
		pe.sendline( "yum -y install " + package )
		result = pe.expect([pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=600)
		if result == 0:
			cb_status( id, node_name, self.FAIL, "package install response timed out" )
			self._log_problem( id, "package install response timed out" )
			return False
		if result != 1:
			cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
			self._log_problem( id, "unknown pexpect return code" )
			return False

		if "Installed: " + package in pe.before:
			cb_status( id, node_name, self.INFO, package + " installed" )
			return True
		if "is already installed." in pe.before:
			cb_status( id, node_name, self.INFO, package + " is already installed" )
			return True
		if "which is newer than" in pe.before:
			cb_status( id, node_name, self.INFO, package + " is already installed and a newer version" )
			return True
		if "No package " + package in pe.before:
			cb_status( id, node_name, self.WARN, "Verify system subscribed to the correct RHN channels" )
			cb_status( id, node_name, self.FAIL, package + " is not available from repository" )
			self._log_problem( id, package + " is not available from repository" )
			return False
		if "could not retrieve mirrorlist" in pe.before:
			cb_status( id, node_name, self.FAIL, "cannot retrieve mirrorlist for repository" )
			self._log_problem( id, "cannot retrieve mirrorlist for repository" )
			return False
		if "You need to be root to perform" in pe.before:
			cb_status( id, node_name, self.WARN, "session does not have root permissions" )
			return True # ????? just for testing
			cb_status( id, node_name, self.FAIL, "session does not have root permissions" )
			self._log_problem( id, "session does not have root permissions" )
			return False
		if "No more mirrors to try" in pe.before:
			cb_status( id, node_name, self.FAIL, "unable get package from available mirrors" )
			self._log_problem( id, "unable get package from available mirrors" )
			return False
		if "Nothing to do" in pe.before:
			cb_status( id, node_name, self.INFO, package + " is already installed" )
			return True
			
		cb_status( id, node_name, self.FAIL, "unable to install " + package )
		if self.DEBUG:
			print "unable to install " + package
			print pe.before

		return False

	def _execute_1cmd(self, pe, id, cmd, cb_status):
		node_name = self.node_list[id]['name']
		pe.sendline( cmd ) 
		result = pe.expect( [pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=600 )
		if result == 0:
			cb_status( id, node_name, self.FAIL, "command (1cmd) response timed out" )
			self._log_problem( id, "command response timed out" )
			self._fail_node( id, cb_finished, cb_progress)
			return False
		if result != 1:
			cb_status( id, node_name, self.FAIL, "unknown (1cmd) pexpect return code" )
			self._log_problem( id, "unknown pexpect return code" )
			self._fail_node( id, cb_finished, cb_progress)
			return False
		return True

	def _execute_steps(self, pe, id, step_list, cb_status, drac=False):
		if self.DEBUG: print "_execute_steps"
		node_name = self.node_list[id]['name']
		fin = False
		jump = False
		jump_target = "start"
		prompt_exp = PEXPECT_PROMPT_EXP
		if drac:
			prompt_exp = PEXPECT_DRAC_PROMPT_EXP
		while not fin:
			for cmd, match_list, label in step_list:
				if jump:
					if jump_target != label:
						continue
					jump = False
				if label == "end":
					if self.DEBUG: print " end"
					fin = True
					break
				# execute cmd
				tmp_cmd = cmd % self.node_list[id]
				cmd = tmp_cmd
				if self.DEBUG: print " sending cmd:" + cmd
				pe.sendline( cmd )
				result = pe.expect([pexpect.TIMEOUT, prompt_exp], timeout=1200)
				if result == 0:
					cb_status( id, node_name, self.FAIL, "command response timed out" )
					self._log_problem( id, "command response timed out" )
					return False
				if result != 1:
					cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
					self._log_problem( id, "unknown pexpect return code" )
					return False
				have_match = False
				for match_str, next_step, sev, msg in match_list:
					if match_str.lower() in pe.before.lower():
						have_match = True
						break
				if not have_match:
					if self.DEBUG: print " expect struct:" + str(pe)
					cb_status( id, node_name, self.FAIL, "unscripted response from command: " + cmd )
					self._log_problem( id, "unscripted response from command: " + cmd )
					return False
				if self.DEBUG: print " match: " + match_str
				if sev != self.TICK:
					cb_status( id, node_name, sev, msg )
				if next_step == "FAIL":
					self._log_problem( id, msg )
					return False
				if next_step == "NEXT":
					continue
				jump_target = next_step
				jump = True
				break

		return True


	def _parse_script( self, pe, id, data, cb_status, cb_progress, script ):
		PARSE_SEP = ";"
		PARSE_LABEL = ":"
		PARSE_JUMP = "jump"
		PARSE_WARN = "warn"
		PARSE_FAIL = "fail"
		PARSE_NEXT = "next"
		PARSE_CODE = "code"
		PARSE_SYNC = "sync"
		PARSE_NODES = "nodes"
		PARSE_NODE1 = "node1"
		PARSE_MGMT = "mgmt"
		PARSE_NODES_FOREACH_NODE = "nodes_foreach_node"
		PARSE_NODES_APP = "nodes_app"
		PARSE_EXIT = "exit"
		PARSE_MATCH_SEP = "|"
		PARSE_MATCH_END = ")"
		PARSE_MATCH_BEGIN = "("

		if self.DEBUG: print "_parse_script"

		prompt_exp = PEXPECT_PROMPT_EXP
		tpe = pe
		spe = self._shell_connect(id, cb_status)
		node_ids = self.get_node_ids()
		on_node_1 = (node_ids[0] == id)

		node_name = self.node_list[id]['name']
		node_ip = self.node_list[id]['private_ipaddr']
		pl = "[" + node_name + "]: "
		if self.DEBUG: print pl, self.node_list[id]

		lex_out = shlex.shlex(script, posix=True )
		jumping = None
		while True:
			token = lex_out.get_token()
			if not token:
				# ??? check if not breaking early
				# end of script
				if jumping:
					print pl, lex_out.error_leader(), "label not found:", jumping
					return False
				if self.DEBUG: print pl, "script complete"
				break

			if jumping:
				if token != PARSE_LABEL:
					# trash reset of line
					while lex_out.get_token() != PARSE_SEP:
						pass
					continue
				label = lex_out.get_token()
				if lex_out.get_token() != PARSE_SEP:
					# ??? error ... malformed seperator
					print pl, lex_out.error_leader(), "malformed label"
					return False
				if label != jumping:
					continue

			if token.lower() == PARSE_LABEL:
				label = lex_out.get_token()
				if lex_out.get_token() != PARSE_SEP:
					print pl,lex_out.error_leader(), "malformed label"
					return False
				continue

			if self.DEBUG: print pl,"SCRIPT:", token

			# EXIT
			if token.lower() == PARSE_EXIT:
				next_token = lex_out.get_token()
				if next_token != PARSE_SEP:
					print pl,lex_out.error_leader(), "malformed exit"
					return False
				if self.DEBUG: print pl,"script complete"
				break

			# JUMP
			if token.lower() == PARSE_JUMP:
				jumping = lex_out.get_token()
				next_token = lex_out.get_token()
				if next_token != PARSE_SEP:
					if lex_out.get_token() != PARSE_SEP:
						# ??? error ... malformed seperator
						print pl,lex_out.error_leader(), "malformed jump"
						return False
				lex_out = shlex.shlex(script, posix=True ) # reset lex_out to begining for reverse loops
				continue

			# FAIL
			if token.lower() == PARSE_FAIL:
				# fail the script... something bad happened (return false???)
				next_token = lex_out.get_token()
				if next_token != PARSE_SEP:
					if lex_out.get_token() != PARSE_SEP:
						# ??? error ... malformed seperator
						print pl,lex_out.error_leader(), "malformed fail"
						return False
				return False

			# CODE
			if token.lower() == PARSE_CODE:
				code = lex_out.get_token()
				next_token = lex_out.get_token()
				if next_token != PARSE_SEP:
					if lex_out.get_token() != PARSE_SEP:
						# ??? error ... malformed seperator
						print pl,lex_out.error_leader(), "malformed code"
						return False
				if self.DEBUG: print pl,".. running code:", repr(code)
				exec code
				continue
        
        
			# SYNC
			if token.lower() == PARSE_SYNC:
				# ??? need to sync between all nodes
				# what if this node is rebooting.. need check point file (node is same as mgmt node)
				if lex_out.get_token() != PARSE_SEP:
					# ??? error ... malformed seperator
					print pl,lex_out.error_leader(), "malformed sync"
					return False

				if not self._check_continue():
					cb_status( id, node_name, self.FAIL, "stopping... another node failed" )
					return False

				cb_status( id, node_name, self.INFO, "waiting for other nodes to synchronize" )

				self.script_lock.acquire()
				self.script_wait_threads += 1
				self.script_lock.release()

				while True:
					time.sleep(1)
					self.script_lock.acquire()
					if self.script_wait_threads == self.script_num_threads:
						self.script_wait_exit = True
						self.script_lock.release()
						break
					if self.script_wait_exit:
						self.script_lock.release()
						break
					self.script_lock.release()

				self.script_lock.acquire()
				self.script_wait_threads -= 1
				self.script_lock.release()

				cb_status( id, node_name, self.INFO, "synchronized" )
				while True:
					self.script_lock.acquire()
					if self.script_wait_threads == 0:
						self.script_lock.release()
						break
					self.script_lock.release()

				self.script_wait_exit = False
				cb_status( id, node_name, self.INFO, "continuing..." )
				
				continue

			# NODE1
			# MGMT
			if token.lower() == PARSE_NODE1 or token.lower() == PARSE_MGMT:
				# skip command line if not on node 1 thread
				if not on_node_1:
					while lex_out.get_token() != PARSE_SEP:
						pass
					continue


			# NODE1
			# MGMT
			# NODES
			if token.lower() == PARSE_NODE1 or token.lower() == PARSE_MGMT or token.lower() == PARSE_NODES or token.lower() == PARSE_NODES_FOREACH_NODE or token.lower() == PARSE_NODES_APP:
				# if MGMT execute local shell.. otherwise remote ssh
				if token.lower() == PARSE_MGMT:
					if self.DEBUG: print pl,"*** setting to local shell"
					tpe = spe
				else:
					if self.DEBUG: print pl,"*** setting to remote shell"
					tpe = pe
				tmp_cmd = lex_out.get_token()
				if lex_out.get_token() != PARSE_MATCH_BEGIN:
					print pl,lex_out.error_leader(), "'(' expected"
					return False

				tmp_token_list = []
				while True:
					tmp_token = lex_out.get_token()
					tmp_token_list.append( tmp_token )
					if tmp_token != PARSE_MATCH_END:
						continue
					msg = lex_out.get_token()
					tmp_token_list.append( msg )
					tmp_token = lex_out.get_token()
					tmp_token_list.append( tmp_token )
					if tmp_token != PARSE_SEP:
						print pl,lex_out.error_leader(), "malformed node command"
						return False
					# put it all back

					while tmp_token_list:
						lex_out.push_token( tmp_token_list.pop() )
					break

				cb_status( id, node_name, self.INFO, msg )
				cluster = []
				current_cluster_data = {}
				mgmt_ipaddr = self._get_local_ip( self.node_list[id]['private_ipaddr'] )
				if self.DEBUG: print "** Management ip is: ", repr(mgmt_ipaddr)
				current_cluster_data = { "me" 	: node_ip,
							"mgmt" 	: mgmt_ipaddr }
				try:
					current_cluster_data["nfs"] = self.nfs_ipaddr
				except:
					pass
				try:
					current_cluster_data["web"] = self.web_ip
				except:
					pass
				try:
					current_cluster_data["samba"] = self.smb_ipaddr
				except:
					pass
				try:
					current_cluster_data["ftp"] = self.ftp_ipaddr
				except:
					pass
				if token.lower() == PARSE_NODES_FOREACH_NODE:
					for node in self.node_list:
						cluster.append(self.node_list[node]['private_ipaddr'])
					if self.DEBUG: print "Vinny DEBUG: cluster: ", repr(cluster)
					for node in cluster:
						current_cluster_data["node"] = node
						command = tmp_cmd % current_cluster_data
						if self.DEBUG: print pl,"doing command:", command
						tpe.sendline(command)
						result = tpe.expect( [pexpect.TIMEOUT, prompt_exp], timeout=1800 )
						if not result: break
				elif token.lower() == PARSE_NODES_APP:
					command = tmp_cmd % current_cluster_data
					if self.DEBUG: print pl,"doing command:", command
					tpe.sendline(command)
					result = tpe.expect( [pexpect.TIMEOUT, prompt_exp], timeout=1800 )
				else:
					command = tmp_cmd % self.node_list[id]
					if self.DEBUG: print pl,"doing command:", command
					tpe.sendline(command)
					result = tpe.expect( [pexpect.TIMEOUT, prompt_exp], timeout=1800 )
				if result == 0:
					cb_status( id, node_name, self.FAIL, "command response timed out" )
					self._log_problem( id, "command response timed out" )
					return False
				if result != 1:
					cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
					self._log_problem( id, "unknown pexpect return code" )
					return False
				have_match = False
				while True:
					next_token = lex_out.get_token()
					if next_token == PARSE_MATCH_END:
						break
					match_str = next_token
					match_cmd = lex_out.get_token()
					if self.DEBUG: print pl,"match_str=", match_str, "match_cmd=", match_cmd
					if not match_str.lower() in tpe.before.lower():
						# no match so go to next match string
						if self.DEBUG: print pl," no match: ", match_str
						while True:
							next_token = lex_out.get_token()
							if next_token == PARSE_MATCH_SEP or next_token == PARSE_MATCH_END:
								if next_token == PARSE_MATCH_END:
									lex_out.push_token( next_token )
								break
						continue
					# have a match ... no do the action
					have_match = True
					if self.DEBUG: print pl," match:", match_str, "match_cmd:", match_cmd
					next_token = lex_out.get_token()
					if self.DEBUG: print pl," next_token", next_token
					if match_cmd == PARSE_NEXT:
						cb_status( id, node_name, self.INFO, next_token )
						break;
					if match_cmd == PARSE_WARN:
						cb_status( id, node_name, self.WARN, next_token )
						break;
					if match_cmd == PARSE_FAIL:
						cb_status( id, node_name, self.FAIL, next_token )
						return False
					if match_cmd == PARSE_CODE:
						code = next_token
						msg = lex_out.get_token()
						if self.DEBUG: print pl,".. running code:", repr(code)
						exec code
						break;
					if match_cmd == PARSE_JUMP:
						jumping = next_token
						msg = lex_out.get_token()
						cb_status( id, node_name, self.INFO, msg )
						break;
					# ??? bad command
					print pl,lex_out.error_leader(), "match command expected"
					return False
					
				if not have_match:
					if self.DEBUG: print pl," expect struct:" + str(pe)
					cb_status( id, node_name, self.FAIL, "unscripted response from command: " + command )
					self._log_problem( id, "unscripted response from command: " + command )
					return False


				while lex_out.get_token() != PARSE_SEP:
					pass
				continue

			# Unknown command
			print pl,lex_out.error_leader(), "unknown command: ", token
			break

		return True


	def _node_shutdown_test( self, id, data, cb_status ):
		node_ip = self.node_list[id]['private_ipaddr']

		# ping the node
		cmd = "ping -c 5 " + node_ip
		pe = pexpect.spawn( cmd )
		pe.expect('5 packets transmitted,\s+([0-9]+)\s+received,\s+(|\+[0-9]+ errors,\s+)([0-9]+)%\s+packet loss, time\s+([0-9]+)ms')
		recv, errors, loss, delay = pe.match.groups()
		if recv == "5":
			# ping is good
			cb_status( id, node_ip, self.INFO, "node still responding, waiting for shutdown..." )
			return False
		elif recv == "0":
			# unable to ping
			cb_status( id, node_ip, self.INFO, "node has shutdown waiting for reboot..." )
			return True
		else:
			# some packets dropped, but still can reach
			cb_status( id, node_ip, self.INFO, "node still responding, waiting for shutdown..." )
			return False
		return True
		

	def _node_recover_test( self, id, data, cb_status ):
		node_ip = self.node_list[id]['private_ipaddr']

		# ping the node
		cmd = "ping -c 5 " + node_ip
		pe = pexpect.spawn( cmd )
		pe.expect('5 packets transmitted,\s+([0-9]+)\s+received,\s+(|\+[0-9]+ errors,\s+)([0-9]+)%\s+packet loss, time\s+([0-9]+)ms')
		recv, errors, loss, delay = pe.match.groups()
		if recv == "5":
			# ping is good
			cb_status( id, node_ip, self.INFO, "node responding to ping, waiting for sshd respawn" )
		elif recv == "0":
			# unable to ping
			return False
		else:
			# some packets dropped, but still can reach
			return False
		return True

	def _node_reboot(self, id, pe, data, cb_status, cb_progress):
		node_name = self.node_list[id]['name']

		if not self._check_continue():
			cb_status( id, node_name, self.FAIL,  "stopping... another node failed" )
			return None


		result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_reboot )
		if not result:
			return None
		pe.close()
		pe = None # to drop connection

		cb_status( id, node_name, self.TICK, "" )
		count = 0
		while not self._node_shutdown_test( id, data, cb_status ):
			count += 1
			cb_status( id, node_name, self.TICK, "" )
			if count > 48:
				cb_status(id, node_name, self.WARN, "System should have shutdown.. please check the system console on " + node_name)
		count = 0
		while not self._node_recover_test( id, data, cb_status ):
			count += 1
			cb_status( id, node_name, self.TICK, "" )
			if count > 48:
				cb_status(id, node_name, self.WARN, "System should have restarted.. please check the system console on " + node_name )
		time.sleep(45)  # ??? wait for sshd to respawn
		count = 0
		while True:
			cb_status( id, node_name, self.TICK, "" )
			pe = self._ssh_connect(id, cb_status, reconnect=True)
			if pe == None:
				time.sleep(2)
				count += 1
				if count > 48:
					cb_status(id, node_name, self.WARN, "System should have started sshd... please check the system console on " + node_name )
				continue
			break
		cb_status( id, node_name, self.INFO, "reconnected to node" )
		return pe


	def _node_ping_test( self, id, data, cb_status, cb_progress ):
		node_ip = self.node_list[id]['private_ipaddr']

		# ping the node
		cmd = "ping -c 5 " + node_ip
		pe = pexpect.spawn( cmd )
		pe.expect('5 packets transmitted,\s+([0-9]+)\s+received,\s+(|\+[0-9]+ errors,\s+)([0-9]+)%\s+packet loss, time\s+([0-9]+)ms')
		recv, errors, loss, delay = pe.match.groups()
		if recv == "5":
			# ping is good
			cb_status( id, node_ip, self.INFO, "good network connection" )
		elif recv == "0":
			# unable to ping
			cb_status( id, node_ip, self.FAIL, "unable to ping node" )
			self._log_problem( id, "unable to ping node" )
			return False
		else:
			# some packets dropped, but still can reach
			cb_status( id, node_ip, self.WARN, "network connection intermittent" )
		return True
		

	def _node_ping(self, id, data, cb_status, cb_finished, cb_progress, first):
		if self.DEBUG: print "_node_ping"
		if not self._node_ping_test( id, data, cb_status, cb_progress ):
			self._fail_node(id, cb_finished, cb_progress)
			return

		# login
		pe = self._ssh_connect(id, cb_status)
		if pe == None:
			cb_status(id, "---", self.FAIL, "unable to ssh to node" )
			self._fail_node(id, cb_finished, cb_progress)
			return
		# -- get hostname if none and put it in node_list if not set
		pe.sendline( 'hostname' )
		pe.readline()
		hostname = pe.readline().strip()
		pe.expect( PEXPECT_PROMPT_EXP )
		if not self.node_list[id].has_key('name'):
			# get hostname from remote and set in list
			cb_status( id, node_ip, self.INFO, "remote hostname: " + hostname )
			self.node_list[id]['name'] = hostname
		else:
			if hostname.lower() == self.node_list[id]['name'].lower():
				if self.DEBUG: print "hostnames names match"
			else:
				cb_status( id, node_ip, self.WARN, "remote hostname (" 
					   + hostname + ") does not match one given (" 
					   + self.node_list[id]['name'] + ")" )
		node_name = self.node_list[id]['name']

		# verify OS version
		result = self._execute_steps( pe, id, self.sl_verify_os, cb_status )
		if not result:
			self._fail_node(id, cb_finished, cb_progress)
			return
		
		self._pass_node(id, cb_finished, cb_progress)
		return

	def _node_configure_all( self, id, data, cb_status, cb_finished, cb_progress, first):
		if self.DEBUG: print "_node_configure_all()"
		try:
			if self.node_list[id].has_key( 'private_ipaddr' ):
				node_ip = self.node_list[id]['private_ipaddr']
			else:
				cb_status( id, "---", self.FAIL, "missing a valid IP address" )
				self._log_problem( id, "not a valide IP address" )
				self._fail_node(id, cb_finished, cb_progress)
				return
			node_ids = self.get_node_ids()
			on_node_1 = (node_ids[0] == id)

			cb_progress( id, 1 )

			do_ipmi    = False
			do_nps     = False  # ??? hack for now
			do_md3000  = False
			do_md3000i = False
			do_el      = False
			do_nfs     = False
			do_web     = False
			do_samba   = False
			do_ftp     = False
			do_virt    = False

			for f,c in self.get_primary_fence():
				if f == self.KEY_IPMI:
					do_ipmi = True
				if "network" in f:  # hack ???
					do_nps = True
			for f,c in self.get_secondary_fence():
				if f == self.KEY_IPMI:
					do_ipmi = True
				if "network" in f:  # hack ???
					do_nps = True

			for f,c in self.get_storage():
				if f == self.KEY_MD3000:
					do_md3000 = True
				if f == self.KEY_MD3000i:
					do_md3000i = True
				if f == self.KEY_EL:
					do_el = True

			for a,c in self.apps_to_install:
				if a == self.KEY_NFS:   do_nfs = True
				if a == self.KEY_WEB:   do_web = True
				if a == self.KEY_SAMBA: do_samba = True
				if a == self.KEY_FTP:   do_ftp = True
				if a == self.KEY_VIRT:  do_virt = True
			


			conf_modules = []
			for item in self.table_keys:
				for mod,c in self.apps_to_install:
					if item == mod:
						conf_modules.append(item)
				for mod,c in self.storage_to_install:
					if item == mod:
						conf_modules.append(item)
				for mod,c in self.primary_fence_to_install:
					if item == mod:
						conf_modules.append(item)
				for mod,c in self.secondary_fence_to_install:
					if item == mod:
						conf_modules.append(item)
				if item == self.KEY_DEFAULT:
					conf_modules.append( item )
					continue

			cb_progress( id, 2 )
			if self.DEBUG: 
				print "**** conf_modules:", conf_modules
				print self.apps_to_install, self.storage_to_install, self.secondary_fence_to_install, self.primary_fence_to_install

			# Execute test for GUI  don't actually execute commands
			if self.TEST or self.APPS_ONLY:
				msgs = [ "verifying node connectivity",
					 "good connection" ]
				if self.drac_config and self.drac_fence:
					msgs.extend([
							"connecting to DRAC",
							"Enabling DRAC telnet support" ])
				msgs.extend([
						"installing repos",
						"updating OS",
						"installing rpms",
						"installing storage drivers",
						"deploying basic cluster",
						"initial configuration complete" ])
				progress = 0
				for msg in msgs:
					if self.FASTTEST or self.APPS_ONLY:
						time.sleep(0.1)
					else:
						time.sleep(random.randrange(2,6))
					progress = progress + 8
					if progress > 100:
						progress = 100
					cb_progress( id, progress )
					cb_status( id, node_ip, self.INFO, msg )
				# test complete
				self._pass_node(id, cb_finished, cb_progress)
				return

			# ping the node
			cb_progress( id, 3 )

			cb_status( id, node_ip, self.INFO, "verifying node connectivity" )
			if not self._node_ping_test( id, data, cb_status, cb_progress ):
				self._fail_node(id, cb_finished, cb_progress)
				return

			cb_progress( id, 4 )
			

			# ??? if drac and using it for fencing then config the DRAC
			# if DRAC isn't used for fencing then telenet support should be left off
			""" ??? will take care of this later in code.. this should probably go away
			if self.drac_config and self.drac_fence:
				if self.DEBUG: print " connecting to node to configure DRAC"
				cb_status( id, node_ip, self.INFO, "Connecting to DRAC" )
				pe = self._ssh_connect(id, cb_status)
				if pe == None:
					cb_status( id, node_ip, self.WARN, "Unable to connect to DRAC will attempt to configure later" )
					# self._fail_node(id, cb_finished, cb_progress)  # ??? don't fail.. instead set IPADDR later
					return
			""" 
			#PROG cb_status( id, node_ip, self.PROG, 10 )
			# login
			pe = self._ssh_connect(id, cb_status)
			if pe == None:
				self._fail_node(id, cb_finished, cb_progress)
				return
			# -- get hostname if none and put it in node_list if not set
			cb_progress(id, 5)
			pe.sendline( 'hostname' )
			pe.readline()
			hostname = pe.readline().strip()
			pe.expect( PEXPECT_PROMPT_EXP )
			if not self.node_list[id].has_key('name'):
				# get hostname from remote and set in list
				cb_status( id, node_ip, self.INFO, "remote hostname: " + hostname )
				self.node_list[id]['name'] = hostname
			else:
				if hostname.lower() == self.node_list[id]['name'].lower():
					if self.DEBUG: print "hostnames names match"
				else:
					cb_status( id, node_ip, self.WARN, "remote hostname (" 
						   + hostname + ") does not match one given (" 
						   + self.node_list[id]['name'] + ")" )
			node_name = self.node_list[id]['name']
			cb_progress(id, 6)

			# testing iSCSI storage via ping

			if do_el or do_md3000i:
				cb_status( id, node_ip, self.INFO, "pinging iSCSI target" )
				for mod,c in self.storage_to_install:
					break
				cmd = "ping -c 5 %(iscsi_target)s" % c

				pe.sendline(cmd)
				pe.expect('5 packets transmitted,\s+([0-9]+)\s+received,\s+(|\+[0-9]+ errors,\s+)([0-9]+)%\s+packet loss, time\s+([0-9]+)ms')
				recv, errors, loss, delay = pe.match.groups()
				if recv == "5":
					# ping is good
					cb_status( id, node_ip, self.INFO, "good iSCSI network connection" )
				elif recv == "0":
					# unable to ping
					cb_status( id, node_ip, self.FAIL, "unable to ping iSCSI target" )
					self._log_problem( id, "unable to ping node" )
					self._fail_node(id, cb_finished, cb_progress)
					return
				else:
					# some packets dropped, but still can reach
					cb_status( id, node_ip, self.WARN, "network connection to iSCSI target intermittent" )
				pe.expect(PEXPECT_PROMPT_EXP) # flushing buffer back to prompt


			cb_status( id, node_name, self.INFO, "excuting script" )
			if self.already_run["cluster.conf"]:
				self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_first_cleanup )
			result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_first )
			if not result:
				self._fail_node(id, cb_finished, cb_progress)
				return

			# Install packages
			cb_progress(id, 35)
			for mod in conf_modules:
				if not self.table_packages.has_key(mod):
					continue
				dict = self.table_packages[mod]
				if dict.has_key(self.NODES):
					for pkg in dict[self.NODES]:
						if self.DEBUG: print "NODES: ", pkg
						if not self._install_package( pe, id, pkg, cb_status ):
							self._fail_node( id, cb_finished, cb_progress )
							return
				if not on_node_1:
					continue
				if dict.has_key(self.MGMT):
					for pkg in dict[self.MGMT]:
						if self.DEBUG: print "MGMT: ", pkg
						cb_status( id, node_name, self.INFO, "Installing cluster managemet package" )
						spe = self._shell_connect(id, cb_status)
						if not self._install_package( spe, id, pkg, cb_status ):
							self._fail_node( id, cb_finished, cb_progress )
							return
						spe.close()
						spe = None
				if dict.has_key(self.NODE1):
					for pkg in dict[self.NODE1]:
						if self.DEBUG: print "NODE1: ", pkg
						if not self._install_package( pe, id, pkg, cb_status ):
							self._fail_node( id, cb_finished, cb_progress )
							return

			if do_md3000 or do_md3000i:
				# Modify MPP to disable LUNRebalance
				result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_mpp_conf)
				if not result:
					self._fail_node(id, cb_finished, cb_progress)
					return

			# Allow firewall ports for cluster communications
			if self.already_run["firewall_cluster"]:
				if self.DEBUG: print "Running cluster firewall cleanup script"
				self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_fw_cluster_cleanup )
			result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_fw_cluster )
			self.already_run["firewall_cluster"] = True
			if not result:
				self._fail_node(id, cb_finished, cb_progress)
				return

			if do_md3000i or do_el:
				if not self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_fw_iscsi ):
					self._fail_node(id, cb_finished, cb_progress)
					return

			# Saving firewall rules
			result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_fw_save )
			if not result:
				self._fail_node(id, cb_finished, cb_progress)
				return


			# ??? print "DRAC DRAC DRAC", self.drac_config, self.drac_fence
			# time.sleep(2)
			# turn on DRAC telnet support
			if self.drac_config:
				cb_status( id, node_name, self.INFO, "Installing DRAC packages" )
				result = self._execute_steps( pe, id, self.sl_install_DRAC, cb_status )
				if not result:
					cb_status( id, node_name, self.FAIL, "Unable to install DRAC packages" )
					self._fail_node(id, cb_finished, cb_progress)
					return
				result = self._execute_steps( pe, id, self.sl_start_DRAC, cb_status )
				if not result:
					cb_status( id, node_name, self.FAIL, "Unable to start DRAC services" )
					self._fail_node(id, cb_finished, cb_progress)
					return
				time.sleep(5)		# Adding sleep to allow all DRAC service to complete starting, as there are many

				if self.drac_fence:
					cb_status( id, node_name, self.INFO, "Enabling telnet for DRAC fencing" )
					result = self._execute_steps( pe, id, self.sl_enable_DRAC_fence, cb_status )
					if not result:
						cb_status( id, node_name, self.FAIL, "Unable to configure DRAC for fencing" )
						self._fail_node(id, cb_finished, cb_progress)
						return

				cb_status( id, node_name, self.INFO, "Configuring DRAC" )
				result = self._execute_steps( pe, id, self.sl_configure_DRAC, cb_status )
				if not result:
					cb_status( id, node_name, self.FAIL, "Unable to configure DRAC" )
					self._fail_node(id, cb_finished, cb_progress)
					return
				cb_status( id, node_name, self.INFO, "Waiting for DRAC reset" )
				time.sleep(60)

			if do_ipmi:
				cb_status( id, node_name, self.INFO, "Configuring IPMI fence" )
				result = self._execute_steps( pe, id, self.sl_enable_IPMI_fence, cb_status )
				if not result:
					cb_status( id, node_name, self.FAIL, "Unable to configure IPMI for fencing" )
					self._fail_node(id, cb_finished, cb_progress)
					return
				cb_status( id, node_name, self.INFO, "Waiting for IPMI reset" )
				time.sleep(10)
				

			cb_progress(id, 39)
			# If Virtualization .. then need to set the correct kernel defaults:
			if do_virt:
				result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_virt_set_kernel )
				if not result:
					self._fail_node(id, cb_finished, cb_progress)
					return
				

			cb_progress(id, 40)
			if not self._check_continue():
				cb_status( id, node_name, self.FAIL, "stopping... another node failed" )
				self._fail_node( id, cb_finished, cb_progress)
				return
					
			if do_virt:

				# ??? need to change /boot/grub/grub.conf to boot xen kernel (default=0)

				pe = self._node_reboot( id, pe, data, cb_status, cb_progress )
				if not pe:
					self._fail_node(id, cb_finished, cb_progress)
					return

			cb_progress(id, 42)

			if not self._check_continue():
				cb_status( id, node_name, self.FAIL,  "stopping... another node failed" )
				self._fail_node( id, cb_finished, cb_progress)
				return

			cb_progress(id, 43)

			cb_progress(id, 47)

			if not self._check_continue():
				cb_status( id, node_name, self.FAIL,  "stopping... another node failed" )
				self._fail_node( id, cb_finished, cb_progress)
				return

			# verify fencing ???
			# deploy cluster.conf

			# lay down conf on each node
			cb_status( id, node_name, self.INFO, "deploying cluster.conf" )
			self.already_run["cluster.conf"] = True
			conf = "cat > /etc/cluster/cluster.conf << EOF\n%(config)s\nEOF\n" % {"config" : self.cluster_conf_basic }
			for line in conf.splitlines():
				pe.sendline( line )
				time.sleep(0.01)
			result = pe.expect( [pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=120 )
			if result == 0: 
				cb_status( id, node_name, self.FAIL, "command response timed out" )
				self._log_problem( id, "command response timed out" )
				self._fail_node(id, cb_finished, cb_progress)
				return False
			if result != 1:
				cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
				self._log_problem( id, "unknown pexpect return code" )
				self._fail_node(id, cb_finished, cb_progress)
				return False

			cb_progress(id, 48)

			# set cluster to start on boot
			cb_status( id, node_name, self.INFO, "enabling ricci on boot" )
			pe.sendline( "chkconfig ricci on" ) 
			result = pe.expect( [pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=120 )
			if result == 0:
				cb_status( id, node_name, self.FAIL, "command response timed out" )
				self._log_problem( id, "command response timed out" )
				self._fail_node(id, cb_finished, cb_progress)
				return False
			if result != 1:
				cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
				self._log_problem( id, "unknown pexpect return code" )
				self._fail_node(id, cb_finished, cb_progress)
				return False
			# storage ????
			
			if not self._check_continue():
				cb_status( id, node_name, self.FAIL,  "stopping... another node failed" )
				self._fail_node( id, cb_finished, cb_progress)
				return

			cb_progress(id, 49)

			if do_el or do_md3000i:
				cb_status( id, node_name, self.INFO, "initializing iSCSI" )
				cb_status( id, node_name, self.INFO, "setting iSCSI initiator name" )
				# edit /etc/iscsi/initiatorname.iscsi
				n_name = self.node_list[id][ "private_ipaddr" ].replace( ".", "-" )
				dict = { "cluster_name":self.cluster_name, "node_name": n_name }
				iscsi_str = "cat > /etc/iscsi/initiatorname.iscsi << EOF\nInitiatorName=iqn.2008-08.%(cluster_name)s:%(node_name)s\nEOF" % dict
				iscsi_str = iscsi_str.replace( "_", "-" )
				if not self._execute_1cmd( pe, id, iscsi_str, cb_status ): return False
				if not self._execute_1cmd( pe, id, "service iscsi start", cb_status ): return False
				if not self._execute_1cmd( pe, id, "chkconfig iscsi on", cb_status ): return False
				for mod,c in self.storage_to_install:
					break
				if do_md3000i:
					pe.sendline( "iscsiadm -m discovery -t sendtargets -p %(iscsi_target)s" % c )
					result = pe.readline()
					result = pe.expect( [pexpect.TIMEOUT, "connection to discovery address [0-9\.]+ failed",
						    PEXPECT_PROMPT_EXP], timeout=300 )
					cb_progress(id, 50)
					if result == 0:
						cb_status( id, node_name, self.FAIL, "command response timed out" )
						self._log_problem( id, "command response timed out" )
						self._fail_node(id, cb_finished, cb_progress)
						return False
					if result == 1:
						cb_status( id, node_name, self.FAIL, "connection to iSCSI target failed." )
						self._log_problem( id, "connection to iSCSI target failed." )
						self._fail_node(id, cb_finished, cb_progress)
						return False
					if result != 2:
						cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
						self._log_problem( id, "unknown pexpect return code" )
						self._fail_node(id, cb_finished, cb_progress)
						return False

				if not self._execute_1cmd( pe, id, "service iscsi start", cb_status ): return False
				if do_el:
					cb_status( id, node_name, self.INFO, "turning on multipathd for EqualLogic support" )
					if not self._execute_1cmd( pe, id, "chkconfig multipathd on", cb_status ): return False
					cb_status( id, node_name, self.INFO, "turning off multipath friendly names and enable multipath support" )
					if not self._execute_1cmd( pe, id, """/bin/sed -i -e 's/^[ \\t]*user_friendly_names[ \\t]*yes/\\tuser_friendly_names no/; s/^[ \\t]*devnode[ \\t]*"\*"/#\\tdevnode "\*"/' /etc/multipath.conf""", cb_status ): return False
					if not self._execute_1cmd( pe, id, "service multipathd stop", cb_status ): return False
					if not self._execute_1cmd( pe, id, "multipath -F", cb_status ): return False
					if not self._execute_1cmd( pe, id, "service multipathd start", cb_status ): return False
					

			cb_progress(id, 51)


			if do_md3000i:
				pass
			if do_md3000:
				pass

			# reboot
			if not self._check_continue():
				cb_status( id, node_name, self.FAIL,  "stopping... another node failed" )
				self._fail_node( id, cb_finished, cb_progress)
				return

			if do_md3000i or do_md3000:		# Vinny: adding explicit checks for MD, as they are the only ones that need reboots for MPP.  Future storage may or may not need reboots
				pe = self._node_reboot( id, pe, data, cb_status, cb_progress )
				if not pe:
					self._fail_node(id, cb_finished, cb_progress)
					return
			else:
				pe.sendline( "service ricci start" )  # Adding service ricci start since someone removed the reboot at the end of this phase for EL
				result = pe.expect( [pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=120 )
				if result == 0:
					cb_status( id, node_name, self.FAIL, "unable to start ricci" )
					self._log_problem( id, "unable to start ricci" )
					self._fail_node(id, cb_finished, cb_progress)
					return False

			cb_progress(id, 80)

			if not self._check_continue():
				cb_status( id, node_name, self.FAIL,  "stopping... another node failed" )
				self._fail_node( id, cb_finished, cb_progress)
				return

			# startup luci ??? fix by loop
			if on_node_1:
				spe = self._shell_connect(id, cb_status)
				# ??? verify we got shell!!
				spe.sendline( "luci_admin init" )
				result = spe.expect( [pexpect.TIMEOUT, 
						      "luci site has been already initialized", 
						      "nter password",
						      "command not found"], timeout=60 )
				if result == 0: 
					cb_status( id, node_name, self.FAIL, "command response timed out" )
					self._log_problem( id, "command response timed out" )
					self._fail_node(id, cb_finished, cb_progress)
					return False
				if result == 1:
					cb_status( id, node_name, self.WARN, "luci was previously initialized" )
					spe.expect( [PEXPECT_PROMPT_EXP] )
					spe.sendline( "service luci stop" )
					spe.expect( [PEXPECT_PROMPT_EXP] )
					spe.sendline( "luci_admin password" )
					result = spe.expect([pexpect.TIMEOUT, "nter new password"], timeout=120)
					if result == 0:
						cb_status( id, node_name, self.FAIL, "luci_init password timed out" )
						self._log_problem( id, "luci_init password timed out" )
						self._fail_node(id, cb_finished, cb_progress)
						return False
					if result != 1:
						cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
						self._log_problem( id, "unknown pexpect return code" )
						self._fail_node(id, cb_finished, cb_progress)
						return False
					result = 2 # ??? cheating just a bit
				if result == 3:
					cb_status( id, node_name, self.FAIL, "luci_admin command not found on mgmt node" )
					self._log_problem( id, "luci_admin command not found on mgmt node" )
					self._fail_node(id, cb_finished, cb_progress)
					return False
				if result != 2:
					cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
					self._log_problem( id, "unknown pexpect return code" )
					self._fail_node(id, cb_finished, cb_progress)
					return False
				spe.sendline( self.cluster_password ) 
				result = spe.expect( [pexpect.TIMEOUT, "onfirm password"], timeout=120 )
				if result == 0: 
					cb_status( id, node_name, self.FAIL, "confirm password prompt timed out" )
					self._log_problem( id, "command response timed out" )
					self._fail_node(id, cb_finished, cb_progress)
					return False
				if result != 1:
					cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
					self._log_problem( id, "unknown pexpect return code" )
					return False
				spe.sendline( self.cluster_password ) 
				result = spe.expect( [pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=300 )
				if result == 0:
					cb_status( id, node_name, self.FAIL, "confirm password done timed out" )
					self._log_problem( id, "command response timed out" )
					self._fail_node(id, cb_finished, cb_progress)
					return False
				if result != 1:
					cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
					self._log_problem( id, "unknown pexpect return code" )
					self._fail_node(id, cb_finished, cb_progress)
					return False
				spe.sendline( "chkconfig luci on" ) 
				result = spe.expect( [pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=120 )
				if result == 0:
					cb_status( id, node_name, self.FAIL, "chkconfig luci timed out" )
					self._log_problem( id, "command response timed out" )
					self._fail_node(id, cb_finished, cb_progress)
					return False
				if result != 1:
					cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
					self._log_problem( id, "unknown pexpect return code" )
					self._fail_node(id, cb_finished, cb_progress)
					return False

				spe.sendline( "service luci restart" )
				spe.expect('Point your web browser to\s+([\S]+)\s+to access luci', timeout=120)
				luci_url = spe.match.groups()[0]
				if self.DEBUG: print "******* luci_url:", luci_url
				self.luci_url = luci_url


			# done with first pass -- storage needs to be probed and verified
			pe.close()
			cb_status( id, node_name, self.INFO, "initial configuration complete" )
			self._pass_node(id, cb_finished, cb_progress)
			return
		except pexpect.EOF:
			if self.DEBUG: print "***BE: Lost connection"
			cb_status( id, "---", self.FAIL, "lost connection" )
			self._log_problem( id, "lost connection" )
			self._fail_node(id, cb_finished, cb_progress)
			raise
			return False
		except:
			print "***BE: got python exception in code."
			traceback.print_exc(file=sys.stdout)
			cb_status( id, "---", self.FAIL, "unhandled python exception in code" )
			self._log_problem( id, "unhandled python exception in code" )
			self._fail_node(id, cb_finished, cb_progress)
			raise
			return False

	def _node_verify_storage( self, id, data, cb_status, cb_finished, cb_progress, first):
		if self.DEBUG: print "_node_verify_storage()"
		try:
			node_ip = self.node_list[id]['private_ipaddr']
			node_ids = self.get_node_ids()
			on_node_1 = (node_ids[0] == id)
			node_name = self.node_list[id]['name']

			local_storage_devs = []

			start_time = time.time()

			do_ipmi    = False
			do_nps     = False  # ??? hack for now
			do_md3000  = False
			do_md3000i = False
			do_el      = False
			do_nfs     = False
			do_web     = False
			do_samba   = False
			do_ftp     = False
			do_virt    = False

			for f,c in self.get_storage():
				if f == self.KEY_MD3000:
					do_md3000 = True
				if f == self.KEY_MD3000i:
					do_md3000i = True
				if f == self.KEY_EL:
					do_el = True


			# Execute test for GUI  don't actually execute commands
			if self.TEST or self.APPS_ONLY:
				msgs = [ "checking configuration",
					 "rebooting",
					 "probing storage",
					 "locating matching scsi_ids",
					 "creating VG",
					 "creating LV",
					 "formating filesystem",
					 "complete" ]
				progress = 0
				for msg in msgs:
					if self.FASTTEST or self.APPS_ONLY:
						time.sleep(0.1)
					else:
						time.sleep(random.randrange(2,6))
					progress = progress + 8
					if progress > 100:
						progress = 100
					cb_progress( id, progress )
					cb_status( id, node_ip, self.INFO, msg )
					if msg == "rebooting" and not self.FASTTEST:
						for X in range(15):
							cb_status( id, node_ip, self.TICK, "" )
							time.sleep(0.5)
				# test complete
				self._pass_node(id, cb_finished, cb_progress)
				return

			# login
			pe = self._ssh_connect(id, cb_status)
			if pe == None:
				self._fail_node(id, cb_finished, cb_progress)
				return

			# restart drivers or system to load the new configuration
			# if not do_el: 			# Vinny adding back in, fixed with switching to /dev/mapper
			if do_md3000i or do_md3000:		# Vinny: switching to explicit checks for MD, as they are the only ones that need reboots for MPP.  Future storage may or may not need reboots
				pe = self._node_reboot( id, pe, data, cb_status, cb_progress )
				if not pe:
					self._fail_node(id, cb_finished, cb_progress)
					return

			if not self._check_continue():
				cb_status( id, node_name, self.FAIL,  "stopping... another node failed" )
				self._fail_node( id, cb_finished, cb_progress)
				return

			if do_el or do_md3000i:
				if do_el:
					result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_iface_check )
					if not result:
						self._fail_node(id, cb_finished, cb_progress)
						return

				cb_status( id, node_name, self.INFO, "login into iSCSI" )
				#if not self._execute_1cmd( pe, id, "service iscsi start", cb_status ): return False
				#if not self._execute_1cmd( pe, id, "chkconfig iscsi on", cb_status ): return False
				for mod,c in self.storage_to_install:
					break
				pe.sendline( "iscsiadm -m discovery -t sendtargets -p %(iscsi_target)s" % c )
				result = pe.readline()
				result = pe.expect( [pexpect.TIMEOUT, "connection to discovery address [0-9\.]+ failed",
					    PEXPECT_PROMPT_EXP], timeout=300 )
				if result == 0:
					cb_status( id, node_name, self.FAIL, "command response timed out" )
					self._log_problem( id, "command response timed out" )
					self._fail_node(id, cb_finished, cb_progress)
					return False
				if result == 1:
					cb_status( id, node_name, self.FAIL, "connection to iSCSI target failed." )
					self._log_problem( id, "connection to iSCSI target failed." )
					self._fail_node(id, cb_finished, cb_progress)
					return False
				if result != 2:
					cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
					self._log_problem( id, "unknown pexpect return code" )
					self._fail_node(id, cb_finished, cb_progress)
					return False

				# ??? handle incorrect IPs for target
				try:
					result = pe.before.splitlines()[0]
					self.target_iqn = result.split()[1]
				except:
					cb_status( id, node_name, self.FAIL, "unable to discover any targets from: %(iscsi_target)s" % c )
					self._log_problem( id, "unable to discover obtain any targets from: %(iscsi_target)s" % c )
					self._fail_node(id, cb_finished, cb_progress)
					return False

				new_parm = {"target_iqn" : self.target_iqn}
				c.update( new_parm )
				if not self._execute_1cmd( pe, id, "service iscsi start", cb_status ): return False
				if self.DEBUG: print repr(c)
				pe.sendline( "iscsiadm -m node -T %(target_iqn)s -p %(iscsi_target)s -l" % c )
				pe.expect( [PEXPECT_PROMPT_EXP] )



			cb_status( id, node_ip, self.INFO, "probing for cluster storage" )
				
			# find /sys/block -maxdepth 1 -name "sd[a-z]" -print
			pe.sendline( 'find /sys/block -maxdepth 1 -name "sd[a-z]" -print' )
			pe.expect( [PEXPECT_PROMPT_EXP] )
			scsi_blk_devs = pe.before.strip().splitlines()[1:-1]
			if self.DEBUG: print scsi_blk_devs
			# cat /sys/block/???/device/module and look for "MD Virtual Disk" for md3k and maybe mkd3ki?  "EQLOGIC" for EL
			# use scsi_id -g -u  -s /block/??? to find ids and match between nodes
			cluster_blk_devs = []
			for blk_dev in scsi_blk_devs:
				pe.sendline( "cat "+blk_dev+"/device/model" )
				pe.expect( [PEXPECT_PROMPT_EXP] )
				buff_model = pe.before
				pe.sendline( "cat "+blk_dev+"/device/vendor" )
				pe.expect( [PEXPECT_PROMPT_EXP] )
				buff_vendor = pe.before
				if self.DEBUG: print node_name, "===" + blk_dev
				if "EQLOGIC".lower() in buff_vendor.lower() or "MD" in buff_model or "NetBSD iSCSI" in buff_model:
					sys_dev = blk_dev.split("/")
					sys_dev.remove('sys')
					sys_dev = "/".join(sys_dev)
					pe.sendline( "cat "+blk_dev+"/size" )
					pe.expect( [PEXPECT_PROMPT_EXP] )
					result = pe.before.splitlines()
					size = result[1]
					pe.sendline( "scsi_id -g -u -s " + sys_dev )
					pe.expect( [PEXPECT_PROMPT_EXP] )
					result = pe.before.splitlines()
					if self.DEBUG: print "result len: ", len(result)
					if len(result) > 2:
						sid = result[1]
					else:
						sid = None
					cluster_blk_devs.append( (blk_dev,sys_dev,size,sid) )
				else:
					if self.DEBUG: print "not cluster storage:", blk_dev, pe.before

			if do_el:
				save_blk_devs = cluster_blk_devs
				cluster_blk_devs = []
				pe.sendline( 'find /dev/mapper -maxdepth 1 -name "[0-9a-f]*[!p][0-9a-f]" -print' )
				pe.expect( [PEXPECT_PROMPT_EXP] )
				scsi_blk_devs = pe.before.strip().splitlines()[1:-1]
				for blk_dev in scsi_blk_devs:
					if self.DEBUG: print node_name, "====" + blk_dev
					sys_dev = ""
					sid = blk_dev.split('/')[-1]
					size = 0
					for  tbdev, tsdev, tsize, tsid in save_blk_devs:
						if sid == tsid:
							size = tsize
							break
					cluster_blk_devs.append( (blk_dev,sys_dev,size,sid) )


			if self.DEBUG: print "+++", repr( cluster_blk_devs )
			# ??? need lock?
			self.cluster_storage_devs_per_node[id] = cluster_blk_devs
			# ??? need unlock?

			# sync up nodes -- everyone done and 1 node can search
			if not self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_sync ):
				cb_status( id, node_ip, self.INFO, "unable to sync nodes" )
				self._fail_node(id, cb_finished, cb_progress)
				return

			# use node 1 to find match
			# search cluster_storage_devs_per_node
			if on_node_1:
				cb_status( id, node_ip, self.INFO, "cross referencing storage probes" )
				ids = self.get_node_ids()
				if self.DEBUG: print repr( self.cluster_storage_devs_per_node )
				list_of_scsi_ids = []
				list_of_no_scsi_ids = []
				#??? cluster_blk_devs.append( (blk_dev,sys_dev,size,sid) )
				for blk_dev, sys_dev, size, sid in self.cluster_storage_devs_per_node[id]:
					if sid:
						list_of_scsi_ids.append( (blk_dev, sys_dev, size, sid) )
					else:
						list_of_no_scsi_ids.append( (blk_dev, sys_dev, size, sid) )
				if self.DEBUG:
					print "HERE IS SCSI IDS"
					print repr( list_of_scsi_ids )
					print "HERE IS NO SCSI IDS"
					print repr( list_of_no_scsi_ids )
				any_match = False
				for blk_dev, sys_dev, size, sid in list_of_scsi_ids:
					node_match = False
					for i in ids:
						dev_match = False
						for mblk_dev, msys_dev, msize, msid in self.cluster_storage_devs_per_node[i]:
							if sid == msid:
								dev_match = True
								break
						if dev_match:
							node_match = True
							continue
						else:
							node_match = False
							break

					if node_match:
						any_match = True
						break

				if any_match:
					cb_status( id, node_ip, self.INFO, "match on SCSI ID accross nodes" )
					# use sys_dev on node_1
				else:
					cb_status( id, node_ip, self.FAIL, "unable to match storage across nodes" )
					self.cluster_storage_devs = []
					self._fail_node(id, cb_finished, cb_progress)
					return

					# here for software iscsi targets that do not do scsi_ids -- not supported
					cb_status( id, node_ip, self.WARN, "unable to find matching scsi ID accross nodes" )
					any_match = False
					for blk_dev, sys_dev, size, sid in list_of_no_scsi_ids:
						node_match = False
						for i in ids:
							dev_match = False
							for mblk_dev, msys_dev, msize, msid in self.cluster_storage_devs_per_node[i]:
								if self.DEBUG: print "looking at ", sys_dev,":", size," ?= ", msys_dev,":", msize
								if size == msize:
									dev_match = True
									break
							if dev_match:
								node_match = True
								continue
							else:
								node_match = False
								break

						if node_match:
							any_match = True
							break
					if any_match:
						cb_status( id, node_ip, self.INFO, "got match on size accross nodes" )
						# use sys_dev on node_1
					else:
						cb_status( id, node_ip, self.FAIL, "unable to match storage across nodes" )
						self.cluster_storage_devs = []
						self._fail_node(id, cb_finished, cb_progress)
						return

				if self.DEBUG: print "** found matching storage:", sys_dev, repr(sid)
				storage_dict = [{ "scsi_id" : sid, 
						  "size": size, 
						  "logical_device": "/dev/vg_cluster/lv_cluster" }]
				if int(size) < 10000000:
					cb_status( id, node_ip, self.FAIL, "clustered storage size is too small, must be larger than 5GB" )
					self._fail_node(id, cb_finished, cb_progress)
					return

				self.cluster_storage_devs = storage_dict
				# do parted ???? of should it go to the next step?
				if do_el:
					dev = blk_dev
				else:
					dev = "/dev/" + sys_dev.split("/")[2]
				self.cs_dev = dev

			# sync up nodes -- other nodes waiting for 1 nodes result
			if not self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_sync ):
				cb_status( id, node_ip, self.INFO, "unable to sync nodes" )
				self._fail_node(id, cb_finished, cb_progress)
				return

			# fail if not match or go on if match
			if not self.cluster_storage_devs:
				cb_status( id, node_ip, self.FAIL, "primary node didn't match storage across nodes" )
				self.cluster_storage_devs = []
				self._fail_node(id, cb_finished, cb_progress)
				return


			# done
			pe.close()
			cb_status( id, node_name, self.INFO, "storage configuration complete" )
			self._pass_node(id, cb_finished, cb_progress)
			return 
		except pexpect.EOF:
			if self.DEBUG: print "***BE: Lost connection"
			cb_status( id, "---", self.FAIL, "lost connection" )
			self._log_problem( id, "lost connection" )
			self._fail_node(id, cb_finished, cb_progress)
			raise
			return False
		except:
			print "***BE: got python exception in code."
			traceback.print_exc(file=sys.stdout)
			cb_status( id, "---", self.FAIL, "unhandled python exception in code" )
			self._log_problem( id, "unhandled python exception in code" )
			self._fail_node(id, cb_finished, cb_progress)
			raise


	def _node_continue_config_all( self, id, data, cb_status, cb_finished, cb_progress, first):
		if self.DEBUG: print "_node_continue_configure_all()"
		try:
			node_ip = self.node_list[id]['private_ipaddr']
			node_ids = self.get_node_ids()
			on_node_1 = (node_ids[0] == id)
			node_name = self.node_list[id]['name']

			do_ipmi    = False
			do_nps     = False  # ??? hack for now
			do_md3000  = False
			do_md3000i = False
			do_el      = False
			do_nfs     = False
			do_web     = False
			do_samba   = False
			do_ftp     = False
			do_virt    = False

			start_time = time.time()


			for f,c in self.get_primary_fence():
				if f == self.KEY_IPMI:
					do_ipmi = True
				if "network" in f:  # hack ???
					do_nps = True
			for f,c in self.get_secondary_fence():
				if f == self.KEY_IPMI:
					do_ipmi = True
				if "network" in f:  # hack ???
					do_nps = True

			for f,c in self.get_storage():
				if f == self.KEY_MD3000:
					do_md3000 = True
				if f == self.KEY_MD3000i:
					do_md3000i = True
				if f == self.KEY_EL:
					do_el = True

			for a,c in self.apps_to_install:
				if a == self.KEY_NFS:   do_nfs = True
				if a == self.KEY_WEB:   do_web = True
				if a == self.KEY_SAMBA: do_samba = True
				if a == self.KEY_FTP:   do_ftp = True
				if a == self.KEY_VIRT:  do_virt = True

			cb_progress( id, 1 )

			# Execute test for GUI  don't actually execute commands
			if self.TEST:
				msgs = [ "verifying storage",
					 "deploying full cluster configuration",
					 "starting cluster",
					 "complete" ]
				progress = 0
				for msg in msgs:
					if self.FASTTEST:
						time.sleep(0.1)
					else:
						time.sleep(random.randrange(2,6))
					progress = progress + 8
					if progress > 100:
						progress = 100
					cb_progress( id, progress )
					cb_status( id, node_ip, self.INFO, msg )
				# test complete
				self._pass_node(id, cb_finished, cb_progress)
				return

			cb_progress( id, 2 )

			# login
			pe = self._ssh_connect(id, cb_status)
			if pe == None:
				self._fail_node(id, cb_finished, cb_progress)
				return

			cb_progress( id, 3 )
			# create mount points
			for key, parm in self.fs_to_create.iteritems():
				cb_status( id, node_name, self.INFO, "creating mount point for clustered storage" )
				pe.sendline( "mkdir -p " + parm[1] )
				pe.expect( [PEXPECT_PROMPT_EXP] )
			

			conf_modules = []
			app_keys = self.app_list.keys()
			for item in app_keys:
				for mod,c in self.apps_to_install:
					if item == mod:
						conf_modules.append(item)
				for mod,c in self.storage_to_install:
					if item == mod:
						conf_modules.append(item)
				for mod,c in self.primary_fence_to_install:
					if item == mod:
						conf_modules.append(item)
				for mod,c in self.secondary_fence_to_install:
					if item == mod:
						conf_modules.append(item)
				if item == self.KEY_DEFAULT:
					conf_modules.append( item )
					continue

			cb_progress( id, 4 )
			if self.DEBUG: 
				print "**** conf_modules:", conf_modules
				print self.apps_to_install, self.storage_to_install, self.secondary_fence_to_install, self.primary_fence_to_install
			# Install packages
			for mod in conf_modules:
				if not self.table_packages.has_key(mod):
					continue
				dict = self.table_packages[mod]
				if dict.has_key(self.NODES):
					for pkg in dict[self.NODES]:
						if self.DEBUG: print "NODES: ", pkg
						if not self._install_package( pe, id, pkg, cb_status ):
							self._fail_node( id, cb_finished, cb_progress )
							return
				if not on_node_1:
					continue
				if dict.has_key(self.MGMT):
					for pkg in dict[self.MGMT]:
						if self.DEBUG: print "MGMT: ", pkg
						spe = self._shell_connect(id, cb_status)
						if not self._install_package( spe, id, pkg, cb_status ):
							self._fail_node( id, cb_finished, cb_progress )
							return
						spe.close()
						spe = None
				if dict.has_key(self.NODE1):
					for pkg in dict[self.NODE1]:
						if self.DEBUG: print "NODE1: ", pkg
						if not self._install_package( pe, id, pkg, cb_status ):
							self._fail_node( id, cb_finished, cb_progress )
							return

			# check storage
			# configure storage (clvm)
			# check mount

			cb_progress( id, 5 )

			if not self._check_continue():
				cb_status( id, node_name, self.FAIL,  "stopping... another node failed" )
				self._fail_node( id, cb_finished, cb_progress)
				return

			# configure firewall for apps
			if do_nfs or do_web or do_samba or do_ftp:
				if self.already_run["firewall_apps"]:
					if self.DEBUG: print "Running apps firewall cleanup script"
					self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_fw_apps_cleanup )
				result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_fw_apps )
				self.already_run["firewall_apps"] = True
				if not result:
					self._fail_node(id, cb_finished, cb_progress)
					return
				cb_progress(id, 6)

				if do_nfs:
					result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_fw_app_nfs )
					if not result:
						self._fail_node(id, cb_finished, cb_progress)
						return
				cb_progress(id, 6)

				if do_web:
					result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_fw_app_web )
					if not result:
						self._fail_node(id, cb_finished, cb_progress)
						return
				cb_progress(id, 6)

				if do_samba:
					result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_fw_app_samba )
					if not result:
						self._fail_node(id, cb_finished, cb_progress)
						return
					pe.sendline("""perl -p -i -e 's/(.*)"/\\1 ip_conntrack_netbios_ns"/ if (/^IPTABLES_MODULES=/ && !/ip_conntrack_netbios_ns/);s/" /"/g;' /etc/sysconfig/iptables-config; depmod -a; modprobe ip_conntrack_netbios_ns""")
					result = pe.expect([pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=600)
					if not result == 1:
						cb_status(id, node_name, self.FAIL, "unable to set ip_conntrack_netbios_ns module")
						self._fail_node(id, cb_finished, cb_progress)
						return

				cb_progress(id, 6)

				if do_ftp:
					result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_fw_app_ftp )
					if not result:
						self._fail_node(id, cb_finished, cb_progress)
						return
					pe.sendline("""perl -p -i -e 's/(.*)"/\\1 ip_conntrack_ftp"/ if (/^IPTABLES_MODULES=/ && !/ip_conntrack_ftp/);s/" /"/g;' /etc/sysconfig/iptables-config; depmod -a; modprobe ip_conntrack_ftp""")
					result = pe.expect([pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=600)
					if not result == 1:
						cb_status(id, node_name, self.FAIL, "unable to set ip_conntrack_ftp module")
						self._fail_node(id, cb_finished, cb_progress)
						return


				cb_progress(id, 6)
				result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_fw_save )
				if not result:
					self._fail_node(id, cb_finished, cb_progress)
					return

			# initialize storage -- partition, clvm
			if self.APPS_ONLY:
				self.cs_dev = "/dev/vg_cluster/lv_cluster"

			dev = self.cs_dev
			self.node_list[id]["cs_dev" ] = dev
			if on_node_1:
				self.node_list[id]["cs_part"] = dev + "1"
				if do_el:
					self.node_list[id]["cs_part"] = dev + "p1"
				self.node_list[id]["cluster_name"] = self.cluster_name
				self.node_list[id]["num_journals"] = len(self.node_list) + 1

			if not self.APPS_ONLY:
				cb_status( id, node_name, self.TICK, "" )
				if self.already_run["lvcreate"]:	# Clean up any previous LVs if needed
					result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_lvcreate_cleanup)
				# Partition script
				if do_md3000 or do_md3000i:
					# These storage types use mpp
					if self.already_run["partition"]:
						self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_partition_mpp_cleanup )
					result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_partition_mpp )
				elif do_el:
					# dm-multipath is handled differently
					if self.already_run["partition"]:
						self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_partition_dm_multipath_cleanup )
					result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_partition_dm_multipath )
				else:
					result = False
				if not result:
					self._fail_node(id, cb_finished, cb_progress)
					return
				self.already_run["partition"] = True
				# Due to multiple changes, split lvcreate into partition/lvcreate
				result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_lvcreate)
				self.already_run["lvcreate"] = True
				if not result:
					self._fail_node(id, cb_finished, cb_progress)
					return

				cb_status( id, node_name, self.TICK, "" )
				cb_status( id, node_name, self.INFO, "adding GFS to fstab" )

				dev_found = False
				for key, parm in self.fs_to_create.iteritems():
					(dev, mp) = parm
					dev_found = True
					break
				if dev_found:
					conf = { "dev" : dev, "mount" : mp }
					if self.DEBUG: print "FSTAB:", repr( conf )
					pe.sendline( "echo '%(dev)s %(mount)s   gfs   debug   0 0' | cat >> /etc/fstab" % conf )
					result = pe.expect( [pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=120 )
					if result == 0: 
						cb_status( id, node_name, self.FAIL, "updating fstab timed out" )
						self._log_problem( id, "command response timed out" )
						self._fail_node(id, cb_finished, cb_progress)
						return
					if result != 1:
						cb_status( id, node_name, self.FAIL, "unknown pexpect return code during fstab update" )
						self._log_problem( id, "unknown pexpect return code during fstab update" )
						self._fail_node(id, cb_finished, cb_progress)
						return

			if on_node_1:
				time.sleep(10)

			time.sleep(10)
			cb_progress( id, 8 )

			# Changed mount code to script, so it can check for errors
			result = self._parse_script( pe, id, data, cb_status, cb_progress, self.scr_mount_gfs)
			if not result:
				self._fail_node(id, cb_finished, cb_progress)
				return
				
			time.sleep(10)

			cb_progress( id, 10 )

			#configure the cluster apps

			subdirs = []
			# WEB
			if do_web:
				subdirs.append( "web/html" )
				subdirs.append( "web/conf" )
				subdirs.append( "web/conf.d" )
				subdirs.append( "web/logs" )
				subdirs.append( "web/run" )
			# SAMBA
			if do_samba:
				subdirs.append( "samba" )
			# FTP
			if do_ftp:
				subdirs.append( "ftp" )
				subdirs.append( "ftp/pub" )
			# NFS
			if do_nfs:  
				# unfortunately it seems that NFS can not be configured 
				# in the cluster environment to be exporting only a 
				# subdir of the GFS volume.
				subdirs.append( "nfs" )

			# create mount points
			if on_node_1:
				cb_progress( id, 13 )
				cb_status( id, node_name, self.INFO, "creating subdirectories for services" )
				for key, parm in self.fs_to_create.iteritems():
					pe.sendline( "mkdir -p " + parm[1] )
					pe.expect( [PEXPECT_PROMPT_EXP] )
					for X in subdirs:
						pe.sendline( "mkdir -p " + parm[1] + "/" + X )
						pe.expect( [PEXPECT_PROMPT_EXP] )
			
			cb_progress( id, 15 )

			# do app specific setup
			# WEB
			if do_web:
				cb_progress( id, 20)
				cb_status( id, node_name, self.INFO, "configuring web service" )
				dict = { "ipaddr" : self.web_ip, "srv_root" : self.web_doc, "doc_root" : self.web_doc + "/html" }
				pe.sendline( "ln -s /etc/httpd/modules %(srv_root)s/modules" % dict )
				pe.expect( PEXPECT_PROMPT_EXP ) 

				# make sure local service control is turned off
				pe.sendline( "chkconfig httpd off" )
				pe.expect( PEXPECT_PROMPT_EXP )

				# save off the original httpd.conf file as httpd.conf.orig
				pe.sendline( r"if [ \! -f /etc/httpd/conf/httpd.conf.orig ]; then mv /etc/httpd/conf/httpd.conf /etc/httpd/conf/httpd.conf.orig; fi" )
				pe.expect( PEXPECT_PROMPT_EXP )

				# make a symbolic link from /etc/httpd/conf/httpd.conf to srv_root/conf/httpd.conf
				pe.sendline( "rm -f /etc/httpd/conf/httpd.conf" )
				pe.expect( PEXPECT_PROMPT_EXP )
				pe.sendline( "ln -s %(srv_root)s/conf/httpd.conf /etc/httpd/conf/httpd.conf " % dict )
				pe.expect( PEXPECT_PROMPT_EXP )

				if on_node_1:  # (only on node1)
					pe.sendline( "cp -fr /var/www/* %(srv_root)s/" % dict )
					pe.expect( PEXPECT_PROMPT_EXP ) 

					# copy /etc/httpd/conf.d to srv_root/conf.d
					pe.sendline( "cp -fr /etc/httpd/conf.d/* %(srv_root)s/conf.d/" % dict )
					pe.expect( PEXPECT_PROMPT_EXP ) 

					# dumping out config file
					conf = "cat > %(srv_root)s/conf/httpd.conf << EOF\n" + http_conf_template + "\nEOF\n"
					if self.DEBUG: 
						print "DEBUG httpd.conf file:"
					for line in conf.splitlines():
						if self.DEBUG: print ">>>", line
						pe.sendline( line % dict )
						time.sleep(0.04)
					pe.expect( PEXPECT_PROMPT_EXP ) 

					# dumping out error page
					conf = "cat > %(srv_root)s/error/noindex.html << EOF\n" + http_readme_template + "\nEOF\n"
					if self.DEBUG: 
						print "DEBUG noindex.html file:"
					for line in conf.splitlines():
						if self.DEBUG: print ">>>", line
						pe.sendline( line % dict )
						time.sleep(0.04)
					pe.expect( PEXPECT_PROMPT_EXP ) 

			# SAMBA
			if do_samba:
				cb_progress( id, 30)
				cb_status( id, node_name, self.INFO, "configuring samba service" )
				dict = { "srv_root" : self.smb_root, "ipaddr" : self.smb_ipaddr, "workgroup": self.smb_workgroup }

				# lay down conf on each node
				conf = "cat > /etc/samba/smb.conf.samba-resource << EOF\n" + samba_conf_template + "\nEOF\n" 
				for line in conf.splitlines():
					pe.sendline( line % dict )
					time.sleep(0.01)
				pe.expect( PEXPECT_PROMPT_EXP ) 


				# make sure local service control is turned off
				pe.sendline( "chkconfig smb off" )
				pe.expect( PEXPECT_PROMPT_EXP )

				if on_node_1:
					# drop a temp file for people to see in the share
					conf = "cat > %(srv_root)s/README << EOF\n" + samba_readme_template + "\nEOF" 
					pe.sendline( conf % dict )
					pe.expect( PEXPECT_PROMPT_EXP ) 

			# FTP
			if do_ftp:
				cb_progress( id, 40)
				cb_status( id, node_name, self.INFO, "configuring ftp service" )
				dict = { "srv_root" : self.ftp_root, "ipaddr" : self.ftp_ipaddr }

				# lay down conf on each node
				conf = "cat > /etc/vsftpd/vsftpd.conf << EOF\n" + ftp_conf_template + "\nEOF\n" 
				for line in conf.splitlines():
					pe.sendline( line % dict )
					time.sleep(0.01)
				pe.expect( PEXPECT_PROMPT_EXP ) 


				# make sure local service control is turned off
				pe.sendline( "chkconfig vsftpd off" )
				pe.expect( PEXPECT_PROMPT_EXP )

				if on_node_1:
					# drop a temp file for people to see in the share
					conf = "cat > %(srv_root)s/README << EOF\n" + ftp_readme_template + "\nEOF" 
					pe.sendline( conf % dict )
					pe.expect( PEXPECT_PROMPT_EXP )

			# NFS
			if do_nfs:
				cb_progress( id, 50)
				cb_status( id, node_name, self.INFO, "configuring nfs service" )
				pass

			pe.sendline( "sync" )
			pe.expect( PEXPECT_PROMPT_EXP )
			time.sleep(5)

			
			# configure clustered apps in cluster.conf
			# redeploy cluster.conf
			if on_node_1:
				cb_progress( id, 60)
				cb_status( id, node_name, self.INFO, "updating cluster.conf" )


				# lay down conf
				cb_progress( id, 65)
				cb_status( id, node_name, self.INFO, "deploying cluster.conf" )
				conf = "cat > /etc/cluster/cluster.conf << EOF\n%(config)s\nEOF\n" % {"config" : self.cluster_conf_full }
				for line in conf.splitlines():
					pe.sendline( line )
					time.sleep(0.01)
				result = pe.expect( [pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=120 )
				if result == 0: 
					cb_status( id, node_name, self.FAIL, "command response timed out" )
					self._log_problem( id, "command response timed out" )
					self._fail_node(id, cb_finished, cb_progress)
					return
				if result != 1:
					cb_status( id, node_name, self.FAIL, "unknown pexpect return code" )
					self._log_problem( id, "unknown pexpect return code" )
					self._fail_node(id, cb_finished, cb_progress)
					return
				# use cli to redeploy cluster.conf
				pe.sendline( "ccs_tool update /etc/cluster/cluster.conf" )
				result = pe.expect( [pexpect.TIMEOUT, PEXPECT_PROMPT_EXP], timeout=120 )
				if result == 0: 
					cb_status( id, node_name, self.FAIL, "cluster.conf update response timed out" )
					self._log_problem( id, "command response timed out" )
					self._fail_node(id, cb_finished, cb_progress)
					return
				if result != 1:
					cb_status( id, node_name, self.FAIL, "unknown pexpect return code from cluster.conf update" )
					self._log_problem( id, "unknown pexpect return code" )
					self._fail_node(id, cb_finished, cb_progress)
					return


			cb_progress( id, 80)
			time.sleep(10) # let the apps come up
			# verify cluster apps ???

			# done
			pe.close()

			cb_status( id, node_name, self.INFO, "node configuration complete" )
			if on_node_1:
				cb_status( id, node_name, self.INFO, "final cluster configuration complete" )
			self._pass_node(id, cb_finished, cb_progress)
			return 
		except pexpect.EOF:
			if self.DEBUG: print "***BE: Lost connection"
			cb_status( id, "---", self.FAIL, "lost connection" )
			self._log_problem( id, "lost connection" )
			self._fail_node(id, cb_finished, cb_progress)
			raise
			return False
		except:
			print "***BE: got python exception in code."
			traceback.print_exc(file=sys.stdout)
			cb_status( id, "---", self.FAIL, "unhandled python exception in code" )
			self._log_problem( id, "unhandled python exception in code" )
			self._fail_node(id, cb_finished, cb_progress)
			raise



	def _build_cluster_confs(self):
		# building basic conf
		doc = Document()

		child_cluster       = doc.createElement("cluster")
		child_fence_daemon  = doc.createElement("fence_daemon")
		child_clusternodes  = doc.createElement("clusternodes")
		child_cman          = doc.createElement("cman")
		child_fencedevices  = doc.createElement("fencedevices")
		child_rm            = doc.createElement("rm")
		child_failoverdomains = doc.createElement("failoverdomains")

		doc.appendChild(child_cluster)

		child_cluster.appendChild(child_fence_daemon)
		child_cluster.appendChild(child_clusternodes)
		child_cluster.appendChild(child_cman)
		child_cluster.appendChild(child_fencedevices)
		child_cluster.appendChild(child_rm)

		# basic cluster attrs
		child_cluster.setAttribute("config_version", "1")
		child_cluster.setAttribute("alias", self.cluster_name)
		child_cluster.setAttribute("name", self.cluster_name)

		# basic fence daemon attrs
		child_fence_daemon.setAttribute("post_fail_delay", "3")
		child_fence_daemon.setAttribute("post_join_delay", "5")
		
		# insert nodes
		num = 0
		for id, node_info in self.node_list.iteritems():
			num = num + 1
			child_node = doc.createElement("clusternode")
			child_clusternodes.appendChild(child_node)
			child_node.setAttribute("name",node_info["private_ipaddr"])
			child_node.setAttribute("nodeid",str(num))
			child_node.setAttribute("votes", "1")
			if self.primary_fence_to_install:
				child_fence = doc.createElement("fence")
				child_node.appendChild(child_fence)
				if not self.secondary_fence_to_install:
					fences = [("1", self.primary_fence_to_install[0])]
				else:
					fences = [("1", self.primary_fence_to_install[0]), ("2", self.secondary_fence_to_install[0])]
				for n, conf in fences:
					if self.DEBUG: print repr(conf)
					t,c = conf
					child_fence_method = doc.createElement("method")
					child_fence_method.setAttribute("name", n)
					child_fence.appendChild(child_fence_method)
					child_fence_device = doc.createElement("device")
					child_fence_method.appendChild(child_fence_device)
					if t == self.KEY_DRAC:
						drac_name = "drac" + str(num)
						child_fence_device.setAttribute("name", drac_name)
						child_fencedevice = doc.createElement("fencedevice")
						child_fencedevices.appendChild(child_fencedevice)
						child_fencedevice.setAttribute("agent", "fence_drac")
						child_fencedevice.setAttribute("name", drac_name)
						child_fencedevice.setAttribute("ipaddr", node_info["drac_ipaddr"])
						child_fencedevice.setAttribute("login", node_info["drac_user"])
						child_fencedevice.setAttribute("passwd", node_info["drac_password"])
					if t == self.KEY_IPMI:
						ipmi_name = "ipmi" + str(num)
						child_fence_device.setAttribute("name", ipmi_name)
						child_fencedevice = doc.createElement("fencedevice")
						child_fencedevices.appendChild(child_fencedevice)
						child_fencedevice.setAttribute("agent", "fence_ipmilan")
						child_fencedevice.setAttribute("name", ipmi_name)
						child_fencedevice.setAttribute("ipaddr", node_info["ipmi_ipaddr"])
						child_fencedevice.setAttribute("login", node_info["ipmi_user"])
						child_fencedevice.setAttribute("passwd", node_info["ipmi_password"])
					if t == self.KEY_WTI:
						child_fence_device.setAttribute("name", "wti")
						child_fence_device.setAttribute("port", "?")
					if t == self.KEY_APC:
						child_fence_device.setAttribute("name", "apc")
						child_fence_device.setAttribute("port", "?")
					if t == self.KEY_NPS:  #??? hack -- should be removed
						child_fence_device.setAttribute("name", "apc")
						child_fence_device.setAttribute("port", "?")
						
			
		# basic cman attrs
		if len( self.node_list ) == 2:
			child_cman.setAttribute("expected_votes", "1")
			child_cman.setAttribute("two_node", "1")

		# insert shared fencedevices
		if self.primary_fence_to_install:
			if not self.secondary_fence_to_install:
				fences = [("1", self.primary_fence_to_install[0])]
			else:
				fences = [("1", self.primary_fence_to_install[0]), ("2", self.secondary_fence_to_install[0])]
			for n, (t, c) in fences:
				if t == self.KEY_WTI:
					child_fencedevice = doc.createElement("fencedevice")
					child_fencedevices.appendChild(child_fencedevice)
					child_fencedevice.setAttribute("agent", "fence_wti")
					child_fencedevice.setAttribute("name", "wti")
					child_fencedevice.setAttribute("ipaddr", c["ipaddr"])
					child_fencedevice.setAttribute("passwd", c["password"])
				if t == self.KEY_APC:
					child_fencedevice = doc.createElement("fencedevice")
					child_fencedevices.appendChild(child_fencedevice)
					child_fencedevice.setAttribute("agent", "fence_apc")
					child_fencedevice.setAttribute("name", "wti")
					child_fencedevice.setAttribute("ipaddr", c["ipaddr"])
					child_fencedevice.setAttribute("login", c["user"])
					child_fencedevice.setAttribute("passwd", c["password"])
				if t == self.KEY_NPS:  #??? hack -- should be removed
					child_fencedevice = doc.createElement("fencedevice")
					child_fencedevices.appendChild(child_fencedevice)
					child_fencedevice.setAttribute("agent", "fence_apc")
					child_fencedevice.setAttribute("name", "apc")
					child_fencedevice.setAttribute("ipaddr", "")
					child_fencedevice.setAttribute("login", "")
					child_fencedevice.setAttribute("passwd", "")


		# insert resource
		child_rm.appendChild(child_failoverdomains)
		
		# save conf
		self.cluster_conf_basic = doc.toprettyxml(indent="    ")

		# building conf with apps
		child_cluster.setAttribute("config_version", "2")
		
		# insert ...
		resources = {}
		for a,c in self.apps_to_install:
			child_service = doc.createElement("service")
			child_rm.appendChild( child_service )
			child_service.setAttribute("autostart", "1")
			child_service.setAttribute("exclusive", "0")
			child_service.setAttribute("recovery","relocate")
			if a == self.KEY_FTP:
				child_service.setAttribute("name","ftp_service")
				resources[ c["ip"] ] = ( "ip", (None))
				child = doc.createElement("ip")
				child_service.appendChild( child )
				child.setAttribute("ref", c["ip"])
				self.ftp_ipaddr = c["ip"]
				c_script = doc.createElement("script")
				child.appendChild( c_script )
				c_script.setAttribute("file", "/etc/init.d/vsftpd")
				c_script.setAttribute("name", "ftp_script")

				gfs_name = "dell" + c["mount"].replace("/", "-")
				parm = (c["dev"], c["mount"])
				resources[ gfs_name ] = ( "fsdev", parm )
				child = doc.createElement("clusterfs")
				child_service.appendChild( child )
				child.setAttribute("fstype", "gfs")
				child.setAttribute("ref", gfs_name )
				self.ftp_root = c["mount"] + "/ftp"

			if a == self.KEY_WEB:
				child_service.setAttribute("name","web_service")
				resources[ c["ip"] ] = ( "ip", (None))
				child = doc.createElement("ip")
				child_service.appendChild( child )
				child.setAttribute("ref", c["ip"])
				self.web_ip = c["ip"]

				c_script = doc.createElement("script")
				child.appendChild( c_script )
				c_script.setAttribute("file", "/etc/init.d/httpd")
				c_script.setAttribute("name", "httpd_script")

				gfs_name = "dell" + c["mount"].replace("/", "-")
				parm = (c["dev"], c["mount"])
				self.web_doc = c["mount"] + "/web"
				resources[ gfs_name ] = ( "fsdev", parm )
				child = doc.createElement("clusterfs")
				child_service.appendChild( child )
				child.setAttribute("fstype", "gfs")
				child.setAttribute("ref", gfs_name )
				ch = doc.createElement("apache")
				# child.appendChild( ch ) ?? shouldn't be adding the apcahe line... 
				# using base script
				ch.setAttribute("config_file","/etc/httpd/conf/httpd.conf")
				ch.setAttribute("name","apache")
				ch.setAttribute("server_root",c["mount"]+"/http")
				ch.setAttribute("shutdown_wait","0")

			if a == self.KEY_SAMBA:
				child_service.setAttribute("name","samba_service")

				resources[ c["ip"] ] = ( "ip", (None))
				child = doc.createElement("ip")
				child_service.appendChild( child )
				child.setAttribute("ref", c["ip"])
				self.smb_ipaddr = c["ip"]

				gfs_name = "dell" + c["mount"].replace("/", "-")
				parm = (c["dev"], c["mount"])
				self.smb_root = c["mount"] + "/samba"
				resources[ gfs_name ] = ( "fsdev", parm )
				child = doc.createElement("clusterfs")
				child_service.appendChild( child )
				child.setAttribute("fstype", "gfs")
				child.setAttribute("ref", gfs_name )

				c_smb = doc.createElement("smb")
				child.appendChild( c_smb )
				c_smb.setAttribute("ref", "samba-resource")
				resources[ "samba-resource" ] = ("samba", c["workgroup"])
				self.smb_workgroup = c["workgroup"]

			if a == self.KEY_NFS:
				child_service.setAttribute("name","nfs_service")
				resources[ c["ip"] ] = ( "ip", (None))
				child = doc.createElement("ip")
				child_service.appendChild( child )
				child.setAttribute("ref", c["ip"])
				self.nfs_ipaddr = c["ip"]
				gfs_name = "dell" + c["mount"].replace("/", "-")
				parm = (c["dev"], c["mount"])
				resources[ gfs_name ] = ( "fsdev", parm )
				child = doc.createElement("clusterfs")
				child_service.appendChild( child )
				child.setAttribute("fstype", "gfs")
				child.setAttribute("ref", gfs_name )
				c_exp = doc.createElement("nfsexport")
				child.appendChild( c_exp )
				c_exp.setAttribute("ref", "nfs_export")
				resources[ "nfs_export" ] = ("export", (None))
				c_cli = doc.createElement("nfsclient")
				c_exp.appendChild( c_cli )
				c_cli.setAttribute("ref", "nfs_client")
				resources["nfs_client"] = ("client", (c["client_opts"], c["clients"], c["mount"]+"/nfs") )
			if a == self.KEY_VIRT:
				child_service.setAttribute("name","virt_service")
				gfs_name = "dell" + c["mount"].replace("/", "-")
				parm = (c["dev"], c["mount"])
				resources[ gfs_name ] = ( "fsdev", parm )
				child = doc.createElement("clusterfs")
				child_service.appendChild( child )
				child.setAttribute("fstype", "gfs")
				child.setAttribute("ref", gfs_name )

		child_res = doc.createElement("resources")
		self.fs_to_create = {}
		child_rm.appendChild( child_res )
		if self.DEBUG: print "RESOURCES:", repr(resources)
		for key, item in resources.iteritems():
			type, parms = item
			if type == "ip":
				child = doc.createElement("ip")
				child.setAttribute("address", key)
				child.setAttribute("monitor_link", "1")
				child_res.appendChild(child)
				continue
			if type == "fsdev":
				child = doc.createElement("clusterfs")
				(dev, mount) = parms
				self.fs_to_create[ key ] = (dev, mount)
				child.setAttribute("device",dev)
				child.setAttribute("force_unmount","0")
				child.setAttribute("fsid",str(random.randrange(10000,99999)))  # set to a random FSID for NFS
				child.setAttribute("fstype","gfs")
				child.setAttribute("mountpoint",mount)
				child.setAttribute("name",key)
				child.setAttribute("self_fence", "0")
				child_res.appendChild( child )
				continue
			if type == "export":
				child = doc.createElement("nfsexport")
				child.setAttribute("name", key)
				child_res.appendChild( child )
				continue
			if type == "client":
				child = doc.createElement("nfsclient")
				(p2, p3, p4) = parms
				child.setAttribute("allow_recover", "0")
				child.setAttribute("name", key)
				child.setAttribute("options", p2)
				child.setAttribute("target", p3)
				child.setAttribute("path", p4)
				child_res.appendChild( child )
				continue
			if type == "samba":
				child = doc.createElement("smb")
				child.setAttribute("name", key)
				child.setAttribute("workgroup", parms)
				child_res.appendChild( child )
				continue
			if type == "script":
				continue
			print "FOUND unsupported TYPE!!!: ", key, type, parms

		# save conf
		self.cluster_conf_full = doc.toprettyxml(indent="    ")


##
##
## main():
##
##


def main():


	def usage():
		print "Usage: dellConfigCluster.py [-h,--help,-D,--debug]"

	try:
		opts, args = getopt.getopt(sys.argv[1:], "hDdt", ["help", "debug","drac","test"])
	except getopt.GetoptError, err:
		# print help information and exit:
		print str(err)
		usage()
		sys.exit(2)

	dcc = DellConfigCluster()

	for i, j in opts:
		if i in ("-D", "--debug"):
			print "enabling debug"
			dcc.set_debug( True )
		elif i in ("-d","--drac"):
			print "configure DRAC requested"
			dcc.set_drac( True )
		elif i in ("-t","--test"):
			dcc.set_test( True )
		elif i in ("-h", "--help"):
			usage()
			sys.exit()
		else:
			assert False, "unhandled option"

#        dcc.add_node( private_ipaddr="172.16.1.26", password="luster", name="node12" )
#        dcc.add_node( private_ipaddr="172.16.1.30", password="cluster", name="node130" )
#        dcc.add_node( private_ipaddr="172.16.2.33", password="cluster", name="node233" )
#        dcc.add_node( private_ipaddr="172.16.2.34", password="cluster", name="node234" )

# virt
	n1 = dcc.add_node( private_ipaddr="192.168.32.51", password="cluster", name="node1" )
	n2 = dcc.add_node( private_ipaddr="192.168.32.52", password="cluster", name="node2" )
	
#	dcc.add_node( private_ipaddr="192.168.32.53", password="cluster", name="node3" )
#	dcc.add_node( private_ipaddr="192.168.32.54", password="cluster", name="node4" )
#	dcc.add_node( private_ipaddr="192.168.32.55", password="cluster", name="node5" )
#	dcc.add_node( private_ipaddr="192.168.32.56", password="cluster", name="node6" )

# dell desk:
#	n1 = dcc.add_node( private_ipaddr="10.9.72.223", password="cluster", name="node1" )
#	n2 = dcc.add_node( private_ipaddr="10.9.72.222", password="cluster", name="node2" )

# dell lab:
#	n1 = dcc.add_node( private_ipaddr="172.16.2.33", password="cluster", name="node233" )
#	n2 = dcc.add_node( private_ipaddr="172.16.2.34", password="cluster", name="node234" )

#	dcc.dump_nodes()

#	dcc.add_node_drac( node_id=n1, ipaddr="172.16.55.33", user="root", password="hobbes", gateway="255.255.0.0")
#	dcc.add_node_drac( node_id=n2, ipaddr="172.16.55.34", user="root", password="hobbes", gateway="255.255.0.0")
	dcc.set_primary_fence( "nps" )
#	dcc.set_secondary_fence( "drac" )


#	dcc.set_storage( 'el', iscsi_target="192.168.100.150" )
	dcc.set_apps_only(True)

	print "apps"
	print dcc.get_apps_list()
	print "storage"
	print dcc.get_storage_list()
	print "fence"
	print dcc.get_fence_list()

	time.sleep(2)


	dcc.set_cluster_name( "turtle" )
#	dcc.nodes_configure_all( dcc.status, dcc.progress, dcc.finished )
#	while dcc.WORKING:
#		print "waiting..."
#		time.sleep(5)
#
#	dcc.PASSED = False  # ?? just a hack for testing
#
#	if not dcc.PASSED:
#		dcc.dump_nodes()
#		print "list of bad nodes: "
#		print dcc.get_bad_nodes()
#		return
#
#	dcc.nodes_verify_storage( dcc.status, dcc.progress, dcc.finished )
#	while dcc.WORKING:
#		print "waiting..."
#		time.sleep(5)
#
#
#	if not dcc.PASSED:
#		dcc.dump_nodes()
#		print "list of bad nodes: "
#		print dcc.get_bad_nodes()
#		return

#	dcc.add_app( dcc.KEY_FTP, ip="172.16.99.101", dev="/dev/vg_cluster/lv_cluster", mount="/nut" )
#
#	dcc.add_app( dcc.KEY_NFS, ip="172.16.99.102", dev="/dev/vg_cluster/lv_cluster", mount="/nut",
#		     clients="172.16.0.0/16", client_opts="rw,soft,sync" )

	dcc.add_app( dcc.KEY_WEB, ip="172.16.99.103", dev="/dev/vg_cluster/lv_cluster", mount="/nut" )

#	dcc.add_app( dcc.KEY_SAMBA, ip="172.16.99.104", dev="/dev/vg_cluster/lv_cluster", mount="/nut",
#		     workgroup="dellunix" )


	dcc.nodes_continue_config_all( dcc.status, dcc.progress, dcc.finished )
	while dcc.WORKING:
		print "waiting..."
		time.sleep(5)

	print "all nodes:"
	dcc.dump_nodes()
	print "list of bad nodes: "
	print dcc.get_bad_nodes()

	return 

if __name__ == "__main__":
	main()

